// \file IjQLStatement.hpp
// \par Description
// Interface class for the jQLStatement class, for use by jEDI drivers.
//
//////////////////////////////////////////////////////////////////////////////////////////
// *ALL* access to jQL objects directly, related to the jQLStatement object *MUST* occur
// within the context of this Interface class.
//////////////////////////////////////////////////////////////////////////////////////////
// This class *MUST ONLY* be changed on the authority of the jEDI Driver Team.
//////////////////////////////////////////////////////////////////////////////////////////
//
// The purpose of this interface class is to 'proxy' all of the jQLStatement object
// access, where used by the jEDI driver.
//
// This is to remove the dependency on the ever-changing jQL classes, and thus prevent
// future crashes should the driver be deployed against a newer version of jBASE.
//
// Primarily, it replicates any necessary methods from the jSQLStatement class, as directly used
// by the jedi driver, and in addition provides the following:
//		o Accessor get_ / set_ methods, in place of direct public member access.
//		o Testing is_ / has_ methods, in place of complex tests against public members.
//		o get_XXXX_asYYY tend to indicate that a cast is involved in the return type.
//			Better here than anywhere else, but still ugly :(
//
//	TEMP_ prefixed methods are temporary, and require refactoring during Phase II
//	They're usually there because the pass back other jQL object types, which are then
//	directly accessed by the driver... For the time being this is hopefully limited to
//	the jQLBaseIIDTest / jQLBaseSelTest 'trees' and their associated 'token' types etc.
//	However, it also looks like some OUTPUT / SORT processor stuff has crept through too :(
//
//	The inclusion of TEMP_ methods, and any newly refactored versions is controlled by the IJQLHEADER_PHASEII_ONLY
//	compile-time symbol. Define this when implementing / testing Phase II; then remove completely, eventually :)
//	With the symbol is not defined (during Phase I), the TEMP_ versions are compiled in, and the 'real' Phase II implementations are not.
//	The opposite becomes true when the symbol is definied.
//
#if !defined(__IjQLStatement_hpp)
#define __IjQLStatement_hpp

#ifdef _WIN32
#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
#endif

#include <jsystem.h>	// For VAR, UINT32 etc.

// Forward declarations of base jBASE jQL types.
class jQLStatement;
class jQLBaseSelTest;
class jQLBaseIIDTest;
class jQLOutputCount;
class jQLOutputSelectList;
class jQLSqlSortProcessor;

// P H A S E  I I // vvv ///////////////////////////////////////////////////////////
// Forward Declare
class IjQLOutputBase;
class IjQLOutputCount;
class IjQLOutputSelectList;
class IjQLBaseTest;
class IjQLBaseSelTest;
class IjQLBaseIIDTest;
class IjQLSqlSortProcessor;

// Can't forward-declare enums, so declare it here
// (Which means it has to be 'removed' from the jedXDB_jQL2SQLdefs.hpp in the Orrible and other drivers...
// enum enTreeTypes;		// Only available in some compilers, non-std ext.
typedef enum IjQL_treeTypes
{
	IJQL_UNKNOWN, IJQL_OR, IJQL_AND, IJQL_WITH
} IjQL_enTreeTypes;
// P H A S E  I I // ^^^ ///////////////////////////////////////////////////////////

/// Interface class for jQLStatement.
class JBUILD_EXPORT IjQLStatement {
public:
	IjQLStatement(jQLStatement* pjQLStatement);
	~IjQLStatement();

	/// Determines if there's a SORT clause, and if it's BY_EXP or BY_EXP_DDSND
	bool is_sortASTType_BY_EXP() const;

	/// Determines if there's an OUTPUT processor
	bool has_OutputProcessor() const;

	/// Determines if there's a SORT processor
	bool has_SortProcessor() const;

	/// Determines if there's a SELECT processor
	bool has_SelectProcessor() const;

	/// Determines if there's an iidAST
	bool has_iidAST() const;

	/// Determines if there's an relationalIID
	bool has_relationalIID() const;

	/// Determines if there's an pass1ItemSource
	bool has_pass1ItemSource() const;

	/// Retreieve SAMPLE value.
	unsigned int get_SampleValue() const;

	/// Executes the jQL statement...
	int Execute(VAR *pSelectList);

	/// Determines if the jQLStatement result type is a COUNT.
	// Still need uses old tricks to access the protected member! :(
	bool is_COUNT_resultType() const;

	// Other accessor methods
	bool get_ValueOutput() const;
	int get_ncol() const;

	// P H A S E  I I //////////////////////////////////////////////////////////////
	// This stuff will come or needs adjusting in Phase II
	// Operations on SelectProcessor and other members.
	// TEMP_ prefix indicates something probably needs doing.
#ifndef IJQLHEADER_PHASEII_ONLY
	/// \todo: Access through/of SelectProcessor?
	UINT32 TEMP_SelectProcessor_GetHighestAttributeUsed() const;

	/// \todo: Direct access of jQLBaseSelTest type through SelectProcessor
	jQLBaseSelTest* TEMP_get_SelectProcessor_SelectionTree() const;

	/// \todo: Direct access of OutputProcessor (at least the casting is in the interface!)
	jQLOutputCount* TEMP_get_OutputProcessor_asOutputCount() const;
	jQLOutputSelectList* TEMP_get_OutputProcessor_asOutputSelectList() const;

	/// \todo: Direct access of SortProcessor (at least the casting is in the interface!)
	jQLSqlSortProcessor* TEMP_get_SortProcessor_asSqlSortProcessor() const;

	/// \todo: Direct access of jQLBaseIIDTest type through pass1ItemSource
	jQLBaseIIDTest* TEMP_get_pass1ItemSource_IIDSelector() const;

	/// \todo: Direct access of tokenType type through SelectionProcessor / SelectionTree
	int TEMP_SelectProcessor_SelectionTree_getNodeType() const;

	/// \todo: Direct access of jQLStatement (for jQLAttrDefn)
	jQLStatement* TEMP_get_jQLStatement_ptr() const;

	/// \todo: Direct access of jQLOutputCount
	jQLOutputCount* TEMP_get_OutputProcessor_asCount() const;
#endif // IJQLHEADER_PHASEII_ONLY

	// P H A S E  I I // vvv ///////////////////////////////////////////////////////////
	IjQLBaseSelTest* replicateTree(jQLBaseSelTest *intest, IjQLBaseSelTest *parent = 0);

	IjQLBaseIIDTest* replicateTree(jQLBaseIIDTest *intest, IjQLBaseIIDTest *parent = 0);

	IjQLBaseSelTest* get_SelectProcessor_SelectionTree() const;

	IjQLBaseIIDTest* get_pass1ItemSource_IIDSelector() const;

	IjQLOutputCount* get_OutputProcessor_asCount() const;

	IjQLOutputSelectList* get_OutputProcessor_asSelectList() const;

	IjQLSqlSortProcessor* get_SortProcessor_asSqlSortProcessor() const;
	
	int SelectProcessor_SelectionTree_getNodeType() const;
	// P H A S E  I I // ^^^ ///////////////////////////////////////////////////////////

protected:

	// private methods
private:
	// Prevent copying...
	IjQLStatement(IjQLStatement& rh);
	IjQLStatement& operator=(IjQLStatement& rh);

private:
	const jQLStatement *_pjQLStatement;

	// P H A S E  I I // vvv ///////////////////////////////////////////////////////////
	IjQLBaseSelTest	*_pSelectProcessorSelectionTree;/// Replicated jQLStatement->SelectProcessor->SelectionTree
	IjQLBaseIIDTest	*_pPass1ItemSourceIIDSelector;	/// Replicated jQLStatement->pass1ItemSource->IIDSelector
	IjQLOutputBase	*_pOutputProcessor;				/// Proxy for jQLStatement->OutputProcessor
	IjQLSqlSortProcessor *_pSortProcessor;			/// Proxy for jQLStatement->SortProcessor
	// P H A S E  I I // ^^^ ///////////////////////////////////////////////////////////
};

#endif // !defined(__IjQLStatement_hpp)
