* Generate output of delimited Values using jQL syntax.
* The delimiter between values can be user-defined, defaults to @TAB.
* Mar 2018 : Daniel Klein

* Initialization

    INCLUDE jQLProperties.h
    EQU space TO CHAR(32)
    EQU dq TO CHAR(34)        ;* double-quote (")
    EQU sq TO CHAR(39)        ;* single-quote (')
    EQU dash TO CHAR(45)
    EQU fb TO CHAR(251)       ;* hex fb = dec 251

#ifdef WINDOWS
    is_windows = @TRUE
    dir_delim = "\"
#else
    is_windows = @FALSE
    dir_delim = "/"
#endif

    command_line = DELETE(SYSTEM(1000), 1)

* Display Help page

    IF command_line = "" OR command_line = "-?" OR UPCASE(command_line) = "-H" THEN
        CRT
        CRT "Desc:   Generates output of delimited values using jQL syntax."
        CRT "        Output can be to a file, stdout or the spooler."
        CRT
        CRT "Syntax: jexport {options} filename {sort-criteria} {selection-critera} {output-specs} {LPTR}"
        CRT
        CRT "Options:"
        CRT
        CRT "     -f filepath"
        CRT "        The full or relative path to where the file is to be generated."
        CRT "        If this option is omitted then output is sent to stdout."
        CRT "        Enclose the path in double-quotes if it contains spaces."
        CRT
        CRT "     -d delimiter"
        CRT "        The character(s) used to separate the values, default is @TAB."
        CRT "        A delimiter can be specified as:"
        CRT
        CRT "           @CSV = Each value is enclosed in quotes and comma separated"
        CRT "           @TAB = Horizontal Tab  char(9)"
        CRT "           @SVM = Sub-Value-Mark  char(252)"
        CRT "           @VM  = Value-Mark      char(253)"
        CRT "           @AM  = Attribute Mark  char(254)"
        CRT "           e.g. -d @VM"
        CRT
        CRT "           Any single character in the range CHAR(33) - CHAR(126), e.g. -d:"
        CRT "           The following characters cannot be used in this manner:"
        CRT "              - (dash)"
        CRT \              " (double-quote)\
        CRT "              ' (single-quote)"
        CRT
        CRT "           \xnn - where 'nn' is a hexadecimal number in the range 0x00 - 0xFF"
        CRT "                  Invalid hex digits defaults to 0x00"
        CRT "                  e.g. -d \xFD"
        CRT
        CRT "     -o Overwrite existing file when the -f option is used, ignored otherwise"
        CRT
        CRT "     -h This page"
        CRT
        CRT "Notes:"
        CRT
        CRT "   * Options are case insensitive and can appear anywhere on the command line."
        CRT
        CRT "   * Column headers are generated from dictionary definitions unless COL-HDR-SUPP is used."
        CRT
        CRT "   * Output will page unless NO.PAGE, NOPAGE, COL-HDR-SUPP or the -f option is used."
        CRT
        CRT "   * Use COUNT-SUPP or NI-SUPP to suppress the 'nnn rows processed' message"
        CRT "     at the end of the report, ignored if the -f option is used."
        CRT
        CRT "   * If no output-specs are specified then the default output for the file is used."
        CRT
        CRT "   * A preceeding select-list will limit the records to process."
        CRT
        CRT "   * The -f option takes precedence over spooler output."
        CRT
        CRT "Examples:"
        CRT
        CRT "JEXPORT PRODUCTS SAMPLE 100"
        CRT 'JEXPORT -D@csv -F .\members.txt -o MEMBERS WITH STATE "OH" BY LASTNAME BY FIRSTNAME LASTNAME FIRSTNAME ZIP MEMBER.NO ID-SUPP'
        CRT "jexport -f /tmp/active_customers -d@vm CUSTOMERS WITH ACTIVITY BY-DSND TOTAL_SALES CUSTNAME TOTAL_SALES"
        STOP
    END

    file_output = @FALSE      ;* output filepath
    overwrite = @FALSE
    delimiter = "@TAB"        ;* default
    csv = @FALSE
    jql_options = 0

    GOSUB process_options
    IF NOT(options_ok) THEN ABORT

    filename = arguments<1>
    DEL arguments<1>
    jql_criteria = CHANGE(arguments, @AM, " ")

* Validate the existence of the file

    OPEN filename TO file THEN CLOSE file  ELSE ABORT 201, filename

* Validate delimiter

    upcase_delimiter = UPCASE(delimiter)
    BEGIN CASE
    CASE upcase_delimiter = "@TAB"
        delimiter = @TAB
    CASE upcase_delimiter = "@CSV"
        delimiter = \","\
        csv  = @TRUE
    CASE upcase_delimiter = "@AM" OR upcase_delimiter = "@FM"
        delimiter = @FM
    CASE upcase_delimiter = "@VM"
        delimiter = @VM
    CASE upcase_delimiter = "@SVM" OR upcase_delimiter = "@SM"
        delimiter = @SVM
    CASE delimiter MATCHES "1X" AND SEQ(delimiter) >= 33 AND SEQ(delimiter) <= 126        ;* a single printable character
    CASE upcase_delimiter[1,2] = "\X" AND LEN(delimiter) = 4
        hex_digits = upcase_delimiter[3,2]
        hex = OCONV(hex_digits,"MCXD")
        delimiter = CHAR(hex)
    CASE 1
        CRT "Invalid delimiter!"
        ABORT
    END CASE

    IF file_output THEN

* Make sure we have a valid parent directory and we have permission to access it

        last_delim_pos = INDEX(output_filepath, dir_delim, COUNT(output_filepath, dir_delim))
        directory_path = output_filepath[1, last_delim_pos-1]
        OPEN directory_path TO dirpath ELSE
            CRT "Directory " : DQUOTE(directory_path) : " does not exist!"
            ABORT
        END
        output_filename = output_filepath[last_delim_pos + 1, -1]
        IF overwrite THEN
            DELETE dirpath, output_filename
        END ELSE
            READ file FROM dirpath, output_filename THEN
                CRT
                CRT "File ":DQUOTE(directory_path:dir_delim:output_filename):" already exists; use the -o option to overwrite."
                ABORT
            END
        END
        CLOSE dirpath

        OPENSEQ output_filepath TO outstream SETTING error_number ON ERROR
            BEGIN CASE
            CASE error_number = "4096"
                error_message = "Network error"
            CASE error_number = "24576"
                error_message = "Permission denied"
            CASE 1
                error_message = "I/O error"
            END CASE
            CRT "Unable to create the output file: ": error_message
            ABORT
        END ELSE
            CREATE outstream ELSE NULL
        END
    END

    selectlist = ""
    READLIST selectlist FROM "" THEN
        jql_options = JQLOPT_USE_SELECT
    END

* Compile the full jQL statement

    request = "retrieve " : filename : " " : jql_criteria
    result_code = JQLCOMPILE(statement, request, jql_options, error_message)
    IF result_code < 0 THEN
        PRINTERR error_message
        ABORT
    END

* Get/Put any desired properties from the compiled jQL statement

    status = JQLGETPROPERTY(column_count, statement, 0, STMT_PROPERTY_OUTCOLS)  ;* get number of column headers
    status = JQLGETPROPERTY(count_supp, statement, 0, STMT_PROPERTY_COUNT_SUPP)
    status = JQLPUTPROPERTY(@TRUE, statement, 0, STMT_PROPERTY_APPLY_OCONV)
    status = JQLGETPROPERTY(lptr, statement, 0, STMT_PROPERTY_LPTR)
    status = JQLGETPROPERTY(nopage, statement, 0, STMT_PROPERTY_NOPAGE)

    IF lptr THEN PRINTER ON

* Construct the column headers

    status = JQLGETPROPERTY(col_hdr_supp, statement, 0, STMT_PROPERTY_COL_HDR_SUPP)
    IF NOT(col_hdr_supp) THEN
        column_headers = ""
        IF status = 0 THEN
            FOR column_number = 1 TO column_count
                status = JQLGETPROPERTY(column_header, statement, column_number, COL_PROPERTY_FORMATTED_HEADING)        ;* gets formatted column headers
                column_headers<column_number> = column_header
            NEXT column_number
        END

        column_header_count = DCOUNT(column_headers, @AM)
        FOR column = 1 TO column_header_count
            column_header = CHANGE(column_headers<column>, fb, @VM)
            column_elements = DCOUNT(column_header, @VM)
            FOR element = 1 TO column_elements
                column_header<1, element> = TRIM(column_header<1, element>, ".", "T")
            NEXT column
            column_headers<column> = CHANGE(TRIM(column_header), @VM, " ")
        NEXT column
        IF csv THEN
            data = CHANGE(column_headers, dq, dq:dq)
            column_headers = dq:column_headers:dq
        END
        column_headers = CHANGE(column_headers, @AM, delimiter)
        IF file_output THEN
            IF is_windows THEN column_headers := @CR
            WRITESEQ column_headers TO outstream ELSE NULL
        END
    END

* Start execution of JQL statement passing the select list if one is being used

    status = JQLEXECUTE(statement, selectlist)
    IF NOT(nopage) AND NOT(col_hdr_supp) AND NOT(file_output) THEN HEADING column_headers
    rows_processed = 0

    LOOP
        status = JQLFETCH(statement, control, data)
    WHILE status >= 0 DO
        DEL data<column_count+1>        ;* Remove redundant trailing attribute
        IF csv THEN
            data = CHANGE(data, dq, dq:dq)
            data = dq:data:dq
        END
        IF delimiter # @AM THEN data = CHANGE(data, @AM, delimiter)
        IF file_output THEN
            IF is_windows THEN data := @CR
            WRITESEQ data TO outstream ELSE NULL
        END ELSE
            PRINT data
        END
        rows_processed++
    REPEAT

    IF file_output THEN
        WEOFSEQ outstream ELSE NULL
        CLOSESEQ outstream
        CRT OCONV(rows_processed, "MD0,") : " rows written to ":output_filepath
    END ELSE
        IF NOT(count_supp) THEN
            PRINT
            PRINT OCONV(rows_processed, "MD0,") : " rows processed."
        END
        IF lptr THEN
            PRINTER OFF
            PRINTER CLOSE
        END
    END
    STOP

process_options:

* Segregate options and arguments

* All options with arguments to the option must be placed in this array
    options_with_arguments = "-D" : @AM : "-F"

    options_ok = @TRUE
    all_options = ""
    arguments = ""  ;* this will contain anything that is not an option
    argument_count = DCOUNT(command_line, @AM)
    FOR attribute = 1 TO argument_count
        this_argument = command_line<attribute>
        IF this_argument[1,1] = "-" THEN
            upcase_option = UPCASE(this_argument)
            LOCATE(upcase_option, options_with_arguments; pos) THEN
                attribute++
                all_options<-1> = this_argument : command_line<attribute>       ;* get the argument to the option
            END ELSE
                all_options<-1> = this_argument   ;* option has no argument or no space between the option and its argument
            END
        END ELSE
            arguments<-1> = this_argument
        END
    NEXT attribute

* Parse options

    options = all_options
    parsed_options = ""
    options = CHANGE(options, @AM, space)
    options = CHANGE(options, sq, dq)
    options = CHANGE(options, dash, @AM)
    DEL options<1>

    num_attr = DCOUNT(options, @AM)
    FOR attr = 1 TO num_attr
        opt = TRIM(options<attr>)
        start_dq_pos = INDEX(opt, dq, 1)
        IF start_dq_pos THEN
            start_option_pos = INDEX(options, @AM, attr-1) + 1
            end_dq_pos = INDEX(options[start_option_pos,-1], dq, 2)
            whole_option_string = TRIM(CHANGE(options[start_option_pos, end_dq_pos], @AM, dash))
            whole_option_string = TRIM(whole_option_string, dq, "A")
            attr += COUNT(whole_option_string, dash)
        END ELSE
            whole_option_string = opt
        END
        opt = whole_option_string[1,1]
        val = TRIM(whole_option_string[2,-1])
        parsed_options<-1> = opt : @VM : val
    NEXT attr
    all_options = parsed_options
    
* Validate options
    
    option_count = DCOUNT(all_options, @AM)
    FOR attribute = 1 TO option_count WHILE options_ok
        opt = all_options<attribute, 1>
        arg = all_options<attribute, 2>
        opt_ok = @TRUE
        arg_ok = @TRUE

        upcase_opt = UPCASE(opt)
        BEGIN CASE
        CASE upcase_opt = "D"
            delimiter = arg
        CASE upcase_opt = "F"
            file_output = @TRUE
            output_filepath = arg
        CASE upcase_opt = "O"
            overwrite = @TRUE
        CASE upcase_opt = "H" OR opt =  "?"
        CASE 1
            opt_ok = @FALSE
        END CASE

        BEGIN CASE
        CASE opt_ok AND arg_ok
        CASE NOT(opt_ok)
            CRT "Unrecognized option: -" : opt
            options_ok = @FALSE
        CASE NOT(arg_ok)
            CRT "Unrecognized argument: " : arg
            options_ok = @FALSE
        END CASE
    NEXT attribute

    RETURN 
