/*
 * This program is used as the interface for Reality-X
 *
 * Copyright (c) 2019 jBASE International. All rights reserved
 *
 * It is loaded and we use pipes to read and write data,
 * so when the user does a READ, the client sends a request
 * to us through the pipe, and then we write back the data
 * to the caller through the write pipe.
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <memory.h>
#include <limits.h>
#include <errno.h>
#include <signal.h>

#include <jportC.h>

#ifdef JBUILD_OPTION_NO_MACROS
#undef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#undef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#else
#undef abs
#include <macros.h>
#endif

#include <jedi.h>
#include <jREALITY.h>
#include <ros/rfc.h>
#include <ros/rlc.h>
#include <ros/rgc.h>


extern char    *optarg;
extern int      optind;
/*
 * This structure and data keeps track of what connections
 * we have made to the Reality database.
 */
struct connections
{
	int             inuse;	/* 1 if this entry is in use, 0 if jfree */
	RFC_ACCOUNT     AccountHandle;	/* The account handle */
	char           *dbase;	/* The name of the database */
	char           *user;	/* The name of the user */
	char           *userpassword;	/* The password of the user */
	char           *account;	/* The name of the account */
	char           *accountpassword;	/* The password of the account */
};
static struct connections *connections_list;
static int      connections_qty;
/*
 * This structure and data keeps track of all opened files.
 */
struct openfiles
{
	int             inuse;	/* 1 if this entry is in use, 0 if jfree */
	RFC_FILE        FileHandle;	/* The file handle returned from the RfcOpenFile() */
	int             connections;	/* The entry in the 'connections' to use. */
	char           *filename;	/* The name of the file */
};
static struct openfiles *openfiles_list;
static int      openfiles_qty;
/*
 * Define the static data areas.
 */
static struct REALITYcmd inbuff;	/* General purpose buffer for receiving messages */
static struct REALITYrtn outbuff;	/* General purpose buffer for sending messages */
static int      RealityDebug;	/* Set if to display the debug error messages */
static char    *ProgramName;	/* Name of this program */
static int      in_wrapup;	/* Shows if (or how often) we have tried to wrap-up */
static int      cache_connect;	/* Set to 1 if to cache the connections so we don't close them */
static char    *version = "$Revision: 1.1.4.2 $";	/* What the current RCS version is */
/*
 * Macro and data to ensure we have the correct account connection
 * set before calling any of the RfcXXX services.
 */
static RFC_ACCOUNT CurrentAccountHandle;
#define	SET_ACCOUNT(newcptr)								\
if (newcptr->AccountHandle != CurrentAccountHandle)			\
{															\
	RfcSetAccount(newcptr->AccountHandle);					\
	CurrentAccountHandle = newcptr->AccountHandle ;			\
}
/*
 * Define the function prototypes.
 */
static void     Wrapup (int exitcode);
static void     SendBaseMessage (int rcode);
static void     SendBytes (void *ptr, int len);
static void     RecieveBytes (void *ptr, int len);
static int      OpenFile (void);
static int      CloseFile (void);
static void     ReadRecord (void);
static int      WriteRecord (void);
static int      DeleteRecord (void);
static int      ClearFile (void);
static int      SelectFile (void);
static int      SelectEndFile (void);
static int      Readnext (void);
static void     StatFile (void);
static int      Lock (void);
static int      DeleteFile (void);
static int      CreateFile (void);
static int      GetFileName (void);
static int      ParseFileConnect (char *inptr, char **filename, char **args, int *connect_number, int *exists_flag);
static int      jcompare (char *string1, char *string2);
static int      jstrdup (char **target, char *source);
static void     SetSignalHandlers (void);
static void     sighandler (int signo);
static void     Usage ();
/*
 */
main (int argc, char *argv[])
{
	char           *p1;
	int             rc, c;
	/*
	 * Set up some signal handlers.
	 */
	SetSignalHandlers ();
	/*
	 * Bit of initialisation malarky.
	 */
	if ((p1 = strrchr (argv[0], '/')) != NULL)
	{
		p1++;
	}
	else
	{
		p1 = argv[0];
	}
	ProgramName = strdup (p1);
	setsid();
	/*
	 * Parse any command line options.
	 */
	while ((c = getopt (argc, argv, "?h")) != -1)
	{
		switch (c)
		{
		case '?':
		case 'h':
			/*
			 * Display the help screen.
			 */
			Usage ();
			exit (0);
		}
	}
	/*
	 * Look for the debug flag and the cache flag.
	 */
	if (getenv ("jREALITY_DEBUG") != NULL)
	{
		RealityDebug = 1;
		fprintf (stderr, "%s: Program started at pid %ld\n", ProgramName, (long) getpid ());
	}
	cache_connect = (getenv ("jREALITY_CACHE_CONNECT") == NULL ? 0 : 1);
	/*
	 * Because of problems with RealitX connect code, we will
	 * permanently enable this.
	 */
	cache_connect = 1;
	/*
	 * Go in a loop getting commands and actioning them.
	 */
	for (;;)
	{
		/*
		 * Get the next command (and wrapup if an error.)
		 */
		RecieveBytes (&inbuff, sizeof (inbuff));
		/*
		 * Default the output buffer.
		 */
		outbuff.operand1 = outbuff.rcode = -1;
		outbuff.buff[0] = '\0';
		/*
		 * Switch on the command.
		 */
		if (RealityDebug)
		{
			inbuff.buff[REALITY_BUFF_SIZE - 1] = '\0';
			fprintf (stderr, "%s: Command ", ProgramName);
			switch (inbuff.cmd)
			{
			case REALITY_CMD_ACK:
				fprintf (stderr, "ACK");
				break;
			case REALITY_CMD_EXIT:
				fprintf (stderr, "EXIT");
				break;
			case REALITY_CMD_OPEN:
				fprintf (stderr, "OPEN");
				break;
			case REALITY_CMD_CLOSE:
				fprintf (stderr, "CLOSE");
				break;
			case REALITY_CMD_READ:
				fprintf (stderr, "READ");
				break;
			case REALITY_CMD_WRITE:
				fprintf (stderr, "WRITE");
				break;
			case REALITY_CMD_DELETE:
				fprintf (stderr, "DELETE");
				break;
			case REALITY_CMD_CLEAR:
				fprintf (stderr, "CLEAR");
				break;
			case REALITY_CMD_SELECT:
				fprintf (stderr, "SELECT");
				break;
			case REALITY_CMD_SELECT_END:
				fprintf (stderr, "SELECT_END");
				break;
			case REALITY_CMD_READNEXT:
				fprintf (stderr, "READNEXT");
				break;
			case REALITY_CMD_JSTAT:
				fprintf (stderr, "JSTAT");
				break;
			case REALITY_CMD_LOCK:
				fprintf (stderr, "LOCK");
				break;
			case REALITY_CMD_CREATE_FILE:
				fprintf (stderr, "CREATE_FILE");
				break;
			case REALITY_CMD_DELETE_FILE:
				fprintf (stderr, "DELETE_FILE");
				break;
			case REALITY_CMD_GET_FILENAME:
				fprintf (stderr, "GET_FILENAME");
				break;
			default:
				fprintf (stderr, "UNKNOWN (%d)", inbuff.cmd);
				break;
			}
			fprintf (stderr, " seen , handle %d\n", inbuff.handle);
		}
		switch (inbuff.cmd)
		{
		case REALITY_CMD_ACK:
			/*
			 * Acknowledge our presence.
			 * This also give us the chance to call the
			 * star up services and check that works.
			 */
			rc = 0;
			if (RgcStartUpServices (rc))
			{
				if (rc == 0)
				{
					rc = EINVAL;
				}
				outbuff.operand1 = 0;
			}
			else
			{
				rc = 0;
				outbuff.operand1 = 0x1234fedc;
			}
			SendBaseMessage (rc);
			break;
		case REALITY_CMD_EXIT:
			/*
			 * Exit this program
			 */
			SendBaseMessage (0);
			Wrapup (0);
			break;
		case REALITY_CMD_OPEN:
			/*
			 * Open a file.
			 */
			SendBaseMessage (OpenFile ());
			break;
		case REALITY_CMD_CLOSE:
			/*
			 * Close a file.
			 */
			SendBaseMessage (CloseFile ());
			break;
		case REALITY_CMD_READ:
			/*
			 * Read in a record.
			 * Note: This function calls SendBaseMessage() itself.
			 */
			ReadRecord ();
			break;
		case REALITY_CMD_WRITE:
			/*
			 * Write a record.
			 */
			SendBaseMessage (WriteRecord ());
			break;
		case REALITY_CMD_DELETE:
			/*
			 * Delete a record.
			 */
			SendBaseMessage (DeleteRecord ());
			break;
		case REALITY_CMD_CLEAR:
			/*
			 * Clear an entire file.
			 */
			SendBaseMessage (ClearFile ());
			break;
		case REALITY_CMD_SELECT:
			/*
			 * Begin the selection of an entire file.
			 */
			SendBaseMessage (SelectFile ());
			break;
		case REALITY_CMD_SELECT_END:
			/*
			 * End the selection of a file.
			 */
			SendBaseMessage (SelectEndFile ());
			break;
		case REALITY_CMD_READNEXT:
			/*
			 * Read the next item id in the select list.
			 */
			SendBaseMessage (Readnext ());
			break;
		case REALITY_CMD_JSTAT:
			/*
			 * Return textual information about the file.
			 * Note: This function calls SendBaseMessage() itself.
			 */
			StatFile ();
			break;
		case REALITY_CMD_LOCK:
			/*
			 * Lock and unlock record locks.
			 */
			SendBaseMessage (Lock ());
			break;
		case REALITY_CMD_DELETE_FILE:
			/*
			 * Delete a file.
			 */
			SendBaseMessage (DeleteFile ());
			break;
		case REALITY_CMD_CREATE_FILE:
			/*
			 * Create a file , or the stub to a file.
			 */
			SendBaseMessage (CreateFile ());
			break;
		case REALITY_CMD_GET_FILENAME:
			/*
			 * Get the name of an existing file.
			 */
			SendBaseMessage (GetFileName ());
			break;
		default:
			/*
			 * Oops , invalid command.
			 */
			SendBaseMessage (EINVAL);
			break;
		}
	}
	/*
	 * We never ever get to this point.
	 */
}

/******************************************************************************
*
* Function:		Wrapup(exitcode)
*
* Description:	Wrapup up the process and exit.
*
* Parameters:	exitcode:		The exit value.
*
* Return Value:	None -- we never return.
*
*****************************************************************************/

static void 
Wrapup (int exitcode)
{
	struct connections *cptr;
	int             i1;
	/*
	 * Show we are in wrapup mode.
	 */
	if ((in_wrapup++) > 10)
	{
		/*
		 * Maybe too many recursive signals, just exit.
		 */
		exit (exitcode > 127 ? 127 : 0);
	}
	/*
	 * Display a trace message.
	 */
	if (RealityDebug)
	{
		if (in_wrapup < 3)
		{
			fprintf (stderr, "%s: Wrapup using exit code %d\n", ProgramName, exitcode);
		}
	}
	/*
	 * Display an error if anything other than error code 0.
	 */
	if (exitcode)
	{
		errno = exitcode;
		perror (ProgramName);
	}
	/*
	 * Do a proper disconnections of all the connections.
	 */
	for (i1 = 0, cptr = connections_list; i1 < connections_qty; i1++, cptr++)
	{
		if (cptr->inuse)
		{
			SET_ACCOUNT (cptr);
			cptr->inuse = 0;
			RfcDisconnect ();
			CurrentAccountHandle = NULL;
		}
	}
	/*
	 * Terminate the entire services.
	 */
	RgcShutDownServices ();
	/*
	 * Sleep to allow time for output pipes to flush.
	 */
	sleep (5);
	/*
	 * Now exit.
	 */
	exit (exitcode > 127 ? 127 : 0);
}

/*****************************************************************************
*
* Function:		SendBaseMessage(rcode)
*
* Description:	Send the base message to the client.
*
* Parameters:	rcode:	The return code of the operation. The caller should 
*						have initialised all other values in the 'outbuff' variable.
*
* Return Value:	None -- we wrapup if an error occurs.
*
****************************************************************************/

static void 
SendBaseMessage (int rcode)
{
	/*
	 * Update the 'rcode' in the send buffer.
	 */
	outbuff.rcode = rcode;
	/*
	 * Display a trace message.
	 */
	if (RealityDebug)
	{
		fprintf (stderr, "%s: Return the base message , rcode %d , operand1 %#.8x\n", ProgramName, outbuff.rcode, outbuff.operand1);
	}
	/*
	 * Send the data using a different (but commonly used) function.
	 */
	SendBytes (&outbuff, sizeof (outbuff));
}

/********************************************************************************
*
* Functions:	SendBytes(ptr , len )
*				RecieveBytes(ptr , len )
*
* Description:	Send or recieve data to/from the client.
*
* Parameters:	ptr:	The buffer to send or recieve
*				len:	The number of bytes to send or recieve.
*
* Return Value:	None, in case of error we wrap up.
*
******************************************************************************/

static void 
SendBytes (void *ptr, int len)
{
	int             i1;
	/*
	 * Send in chunks at a time. This way, if we get interrupted
	 * or anything, we can re-try the operation.
	 */
	while (len > 0)
	{
		errno = 0;
		i1 = write (1, ptr, len);
		if (i1 != len && errno != EINTR)
		{
			Wrapup ((errno == 0 ? EINVAL : errno));
		}
		if (i1 > 0)
		{
			len -= i1;
			ptr = (((char *) ptr) + i1);
		}
	}
}

static void 
RecieveBytes (void *ptr, int len)
{
	int             i1;
	/*
	 * Recieve in chunks at a time. This way, if we get interrupted
	 * or anything, we can re-try the operation.
	 */
	while (len > 0)
	{
		errno = 0;
		i1 = read (0, ptr, len);
		if (i1 != len && errno != EINTR)
		{
			/*
			 * If the errno is zero and number of bytes
			 * recieved is zero, we can assume it is a normal wrapup.
			 */
			Wrapup ((errno == 0 ? (i1 == 0 ? 0 : EINVAL) : errno));
		}
		if (i1 > 0)
		{
			len -= i1;
			ptr = (((char *) ptr) + i1);
		}
	}
	return;
}

/**************************************************************************************
*
* Function:		OpenFile()
*
* Description:	Perform a file open.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the file was opened, any other it the error number.
*
**************************************************************************************/

static int 
OpenFile ()
{
	struct openfiles *optr, *ofound;
	RFC_FILE        filehandle;
	char           *filename, *args;
	int             i1, rc, connect_number, exists_flag;
	/*
	 * Display any verbose debug information.
	 */
	if (RealityDebug)
	{
		fprintf (stderr, "%s: Open file with specification %s\n", ProgramName, inbuff.buff);
	}
	/*
	 * Parse the input specification and try to open
	 * the connection it references.
	 */
	if ((rc = ParseFileConnect (inbuff.buff, &filename, &args, &connect_number, &exists_flag)) != 0)
	{
		return rc;
	}
	/*
	 * See if we can actually open the file !
	 */
	if ((rc = RfcOpenFile (filename, &filehandle)) != 0)
	{
		return rc;
	}
	/*
	 * Allocate an entry in the 'openfiles_list' to accomodate.
	 */
	for (i1 = 0, optr = openfiles_list, ofound = NULL; i1 < openfiles_qty; i1++, optr++)
	{
		if (!optr->inuse)
		{
			ofound = optr;
			break;
		}
	}
	if (ofound == NULL)
	{
		if ((openfiles_list = (openfiles_list == NULL ? jmalloc (sizeof (struct openfiles)) : jrealloc (openfiles_list, sizeof (struct openfiles) * (openfiles_qty + 1)))) == NULL)
		{
			return errno;
		}
		ofound = openfiles_list + openfiles_qty++;
	}
	/*
	 * Fill in the members in the found list.
	 */
	ofound->FileHandle = filehandle;	/* Add the recently opened file handle */
	ofound->connections = connect_number;	/* The entry number of the connection */
	if ((ofound->filename = strdup (filename)) == NULL)	/* Add the name of the file we opened */
	{
		return errno;
	}
	ofound->inuse = 1;	/* Show it is in use */
	/*
	 * Set the 'operand1' to show what opened file descriptor we allocated.
	 */
	outbuff.operand1 = ofound - openfiles_list;
	/*
	 * Errrr That's it !
	 */
	if (RealityDebug)
	{
		fprintf (stderr, "%s: File %s opened okay to handle %d , database '%s', connection number %d\n", ProgramName, filename, outbuff.operand1, (connections_list + ((openfiles_list + outbuff.operand1)->connections))->dbase, (openfiles_list + outbuff.operand1)->connections);
	}
	return 0;
}

/**************************************************************************************
*
* Function:		CloseFile()
*
* Description:	Perform a file close.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the file was closed, any other it the error number.
*
**************************************************************************************/

static int 
CloseFile ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             i1, connections_number, found;
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	/*
	 * Simply call Reality to close the file.
	 */
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	RfcClose (optr->FileHandle);	/* Actually close the file */
	optr->inuse = 0;	/* Show no longer in use */
	jfree (optr->filename);	/* Release the space used by the file name */
	/*
	 * Count the number of current files that are using the same
	 * connection. If there are none left, we can terminate that connection.
	 *
	 * For speed though the user may have chosen to ignore this.
	 */
	/*
	 * Note: This bit is permanently disabled. There appears to be a 
	 * bug in the RealityX drivers that means with lots of connections
	 * it eventually gives up. Can't find the bug, but I know
	 * if I leave the connection open then I'm okay.
	 */
	if (!cache_connect)
	{
		for (i1 = 0, found = 0, optr = openfiles_list; i1 < openfiles_qty; i1++, optr++)
		{
			if (optr->inuse && optr->connections == connections_number)
			{
				/*
				 * Somebody else is already using this connection.
				 */
				found++;
			}
		}
		if (!found)
		{
			if (RealityDebug)
			{
				fprintf (stderr, "%s: Disconnect from database %s , connection number %d\n", ProgramName, cptr->dbase, cptr - connections_list);
			}
			/*
			 * Nobody else is using this connection, so get rid of it.
			 */
			cptr->inuse = 0;
			RfcDisconnect ();
			CurrentAccountHandle = NULL;
			jfreezero ((char **) &cptr->dbase);
			jfreezero ((char **) &cptr->user);
			jfreezero ((char **) &cptr->userpassword);
			jfreezero ((char **) &cptr->account);
			jfreezero ((char **) &cptr->accountpassword);
		}
	}
	/*
	 * Return to caller with successful completion code.
	 */
	return 0;
}

/**************************************************************************************
*
* Function:		ReadRecord()
*
* Description:	Read a record from a file.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	None.
*
**************************************************************************************/

static void 
ReadRecord ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             i1, rc, rlen, buffused;
	char            buffer[4096], *p1, *rdata;
	if (RealityDebug)
	{
		fprintf (stderr, "%s: Read record , key '", ProgramName);
		fwrite (inbuff.buff, 1, inbuff.operand1, stderr);
		fprintf (stderr, "' , flags %#.8x , field %d\n", inbuff.operand2, inbuff.operand3);
	}
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		SendBaseMessage (EINVAL);
		return;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		SendBaseMessage (EINVAL);
		return;
	}
	/*
	 * Set the right account pointer for this operation.
	 */
	cptr = connections_list + optr->connections;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	/*
	 * There are up to 4 different reality functions we might have to call.
	 *   READ
	 *   READU
	 *   READV
	 *   READVU
	 * They have chosen to do it with 4 functions instead of 1 and some flags.
	 * Ho hum, if I'd been working there I'd have given Barkus a good kicking
	 * for letting this go through.
	 *
	 * The length of the key is passed in 'operand1'.
	 * The flags field is passed in 'operand2'.
	 * The field number (attribute number) is passed in 'operand3'.
	 */
	/*
	 * Set the options depending on whether we have a LOCKED clause or not.
	 */
	if (inbuff.operand2 & JEDI_RECORD_LOCKRECORD)
	{
		RfcSetLockMode (RFC_OPT_HOLD);
	}
	else if (inbuff.operand2 & JEDI_RECORD_LOCKRECORD_NOWAIT)
	{
		RfcSetLockMode (RFC_OPT_NO_WAIT);
	}
	else
	{
		RfcSetLockMode (RFC_OPT_NONE);
	}
	/*
	 * Decide what function to call.
	 */
	if (inbuff.operand2 & JEDI_RECORD_FIELDREADWRITE)
	{
		/*
		 * Either READV or READVU.
		 */
		if (inbuff.operand2 & (JEDI_RECORD_LOCKRECORD | JEDI_RECORD_LOCKRECORD_NOWAIT))
		{
			/*
			 * Must be a READVU.
			 */
			rc = RfcLockReadAttr (
						     optr->FileHandle,	/* The file handle */
						     inbuff.buff,	/* The item id */
						     inbuff.operand1,	/* The item id length */
						     inbuff.operand3,	/* The attribute number to return */
						     buffer,	/* Where to return the data to */
						     sizeof (buffer),	/* How big the return buffer is */
						     &rlen);	/* Perhaps the real length */
		}
		else
		{
			/*
			 * Must be a READV.
			 */
			rc = RfcReadAttr (
						 optr->FileHandle,	/* The file handle */
						 inbuff.buff,	/* The item id */
						 inbuff.operand1,	/* The item id length */
						 inbuff.operand3,	/* The attribute number to return */
						 buffer,	/* Where to return the data to */
						 sizeof (buffer),	/* How big the return buffer is */
						 &rlen);	/* Perhaps the real length */
		}
	}
	else
	{
		/*
		 * Either READ or READU.
		 */
		if (inbuff.operand2 & (JEDI_RECORD_LOCKRECORD | JEDI_RECORD_LOCKRECORD_NOWAIT))
		{
			/*
			 * Must be a READU.
			 */
			rc = RfcLockRead (
						 optr->FileHandle,	/* The file handle */
						 inbuff.buff,	/* The item id */
						 inbuff.operand1,	/* The item id length */
						 buffer,	/* Where to return the data to */
						 sizeof (buffer),	/* How big the return buffer is */
						 &rlen);	/* Perhaps the real length */
		}
		else
		{
			/*
			 * Must be a READ.
			 */
			rc = RfcRead (
					     optr->FileHandle,	/* The file handle */
					     inbuff.buff,	/* The item id */
					     inbuff.operand1,	/* The item id length */
					     buffer,	/* Where to return the data to */
					     sizeof (buffer),	/* How big the return buffer is */
					     &rlen);	/* Perhaps the real length */
		}
	}
	/*
	 * Decide what to do based on the return code.
	 */
	switch (rc)
	{
	case 0:
		/*
		 * Well , what do you know , it all worked !
		 */
		rdata = buffer;	/* Address of where the data can be found */
		break;
	case RFE_NOITEM:
		/*
		 * Item does not exist.
		 */
		SendBaseMessage (ENOENT);
		return;
	case RFE_NOATTR:
		/*
		 * The item exists but not the attribute.
		 */
		rdata = buffer;
		rlen = 0;
		break;
	case RFE_LOCKED:
		/*
		 * Item is already locked and we used the NOWAIT bit.
		 */
		SendBaseMessage (JEDI_ERRNO_LOCK_TAKEN);
		return;
	case RFE_READEXCEED:
		/*
		 * Oh bugger , we need to call that crappy RfcReadRest function.
		 * Allocate a buffer big enough to take what we have so far.
		 */
		if ((rdata = jmalloc (sizeof (buffer))) == NULL)
		{
			SendBaseMessage (errno);
			return;
		}
		memcpy (rdata, buffer, sizeof (buffer));	/* Copy initial data */
		rlen = sizeof (buffer);		/* Amount of data used so far */
		/*
		 * Call 'RfcReadRest' until no more.
		 */
		while (rc == RFE_READEXCEED)
		{
			/*
			 * Get the next batch of data.
			 */
			rc = RfcReadRest (optr->FileHandle, buffer, sizeof (buffer), &buffused);
			if (rc != 0 && rc != RFE_READEXCEED)
			{
				SendBaseMessage (rc);
				return;
			}
			/*
			 * Re-allocate some extra space.
			 */
			if ((rdata = jrealloc (rdata, buffused + rlen)) == NULL)
			{
				SendBaseMessage (errno);
				return;
			}
			/*
			 * Copy the data just read in.
			 */
			memcpy (rdata + rlen, buffer, buffused);
			rlen += buffused;
		}
		break;
	default:
		/*
		 * Indeterminate error.
		 */
		SendBaseMessage (rc);
		return;
	}
	/*
	 * At this point in time we have data to return to the user.
	 * The variable 'rdata' points to the address of the item.
	 * The variable 'rlen' shows the length of the item.
	 *
	 * We should now attempt to return to the caller as much as we 
	 * can in the basic structure, and then follow that with the
	 * rest of the item we couldn't fit in.
	 */
	outbuff.operand1 = rlen;	/* Total length of the item */
	i1 = min (rlen, REALITY_BUFF_SIZE);	/* The length to fit into this buffer */
	p1 = rdata;		/* Take a copy of the data to return */
	memcpy (outbuff.buff, p1, i1);	/* Copy the first bit of the item, or ALL the item ! */
	SendBaseMessage (0);
	/*
	 * Send any remainder of the message.
	 */
	rlen -= i1;		/* Number of bytes still to send */
	p1 += i1;		/* Address of next data byte to send */
	while (rlen > 0)
	{
		i1 = min (REALITY_BLOCK_SIZE, rlen);	/* Number of bytes to transmit this time */
		SendBytes (p1, i1);	/* Send this block of data */
		rlen -= i1;	/* Adjust number of bytes left to send */
		p1 += i1;	/* Adjust the pointer to the next data to send */
	}
	/*
	 * Release any temporary workspace.
	 */
	if (rdata != buffer)
	{
		jfree (rdata);
	}
	/*
	 * Well , it appears everything has gone to plan !
	 */
	return;
}

/**************************************************************************************
*
* Function:		WriteRecord()
*
* Description:	Write a record to a file.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the operation worked, any other is the error number.
*
**************************************************************************************/

static int 
WriteRecord ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             i1, rc, rlen;
	char           *p1, *dataptr;
	if (RealityDebug)
	{
		fprintf (stderr, "%s: Write record , key '", ProgramName);
		fwrite (inbuff.buff, 1, inbuff.operand1, stderr);
		fprintf (stderr, "' , flags %#.8x , field %d , data length %d\n", inbuff.operand2, inbuff.operand3, inbuff.operand4);
	}
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	/*
	 * Set the right account pointer for this operation.
	 */
	cptr = connections_list + optr->connections;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	/*
	 * There are up to 4 different reality functions we might have to call.
	 *   WRITE
	 *   WRITEU
	 *   WRITEV
	 *   WRITEVU
	 *
	 * The length of the key is passed in 'operand1'.
	 * The flags field is passed in 'operand2'.
	 * The field number (attribute number) is passed in 'operand3'.
	 * The length of the data is passed in 'operand4'.
	 */
	/*
	 * Get the rest of the data we might need.
	 */
	if ((dataptr = jmalloc (inbuff.operand4 + 10)) == NULL)
	{
		return errno;
	}
	rlen = inbuff.operand4;	/* Number of bytes still to recieve */
	p1 = dataptr;		/* Address of next data byte to recieve */
	while (rlen > 0)
	{
		i1 = min (REALITY_BLOCK_SIZE, rlen);	/* Number of bytes to recieve this time */
		RecieveBytes (p1, i1);	/* Recieve this block of data */
		rlen -= i1;	/* Adjust number of bytes left to recieve */
		p1 += i1;	/* Adjust the pointer to the next data to recieve */
	}
	/*
	 * Decide what function to call.
	 */
	if (inbuff.operand2 & JEDI_RECORD_FIELDREADWRITE)
	{
		/*
		 * Either WRITEV or WRITEVU.
		 */
		if (inbuff.operand2 & (JEDI_RECORD_LOCKRECORD | JEDI_RECORD_LOCKRECORD_NOWAIT))
		{
			/*
			 * Must be a WRITEVU.
			 */
			rc = RfcWriteAttr (
						  optr->FileHandle,	/* The file handle */
						  inbuff.buff,	/* The item id */
						  inbuff.operand1,	/* The item id length */
						  inbuff.operand3,	/* The attribute number to write */
						  dataptr,	/* The data to write out */
						  inbuff.operand4);	/* The number of bytes in 'dataptr' */
		}
		else
		{
			/*
			 * Must be a WRITEV.
			 */
			rc = RfcWriteAttrUnlock (
							optr->FileHandle,	/* The file handle */
							inbuff.buff,	/* The item id */
							inbuff.operand1,	/* The item id length */
							inbuff.operand3,	/* The attribute number to write */
							dataptr,	/* The data to write out */
							inbuff.operand4);	/* The number of bytes in 'dataptr' */
		}
	}
	else
	{
		/*
		 * Either WRITE or WRITEU.
		 */
		if (inbuff.operand2 & (JEDI_RECORD_LOCKRECORD | JEDI_RECORD_LOCKRECORD_NOWAIT))
		{
			/*
			 * Must be a WRITEU.
			 */
			rc = RfcWrite (
					      optr->FileHandle,		/* The file handle */
					      inbuff.buff,	/* The item id */
					      inbuff.operand1,	/* The item id length */
					      dataptr,	/* The data to write out */
					      inbuff.operand4);		/* The number of bytes in 'dataptr' */
		}
		else
		{
			/*
			 * Must be a WRITE.
			 */
			rc = RfcWriteUnlock (
						    optr->FileHandle,	/* The file handle */
						    inbuff.buff,	/* The item id */
						    inbuff.operand1,	/* The item id length */
						    dataptr,	/* The data to write out */
						    inbuff.operand4);	/* The number of bytes in 'dataptr' */
		}
	}
	/*
	 * Release any temporary workspace.
	 */
	jfree (dataptr);
	/*
	 * Well , it appears everything has gone to plan !
	 */
	return rc;
}

/**************************************************************************************
*
* Function:		DeleteRecord()
*
* Description:	Delete a record.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the record was deleted, any other it the error number.
*
**************************************************************************************/

static int 
DeleteRecord ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             rc, connections_number;
	if (RealityDebug)
	{
		fprintf (stderr, "%s: Delete record , key '", ProgramName);
		fwrite (inbuff.buff, 1, inbuff.operand1, stderr);
		fprintf (stderr, "'\n");
	}
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	/*
	 * Simply call Reality to delete a record.
	 */
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	rc = RfcDelete (optr->FileHandle, inbuff.buff, inbuff.operand1);	/* Delete the record */
	/*
	 * Convert any return codes from Reality to what jBASE expects.
	 */
	switch (rc)
	{
	case RFE_NOITEM:
		/*
		 * The item doesn't exist so it can't be deleted.
		 */
		rc = ENOENT;
		break;
	}
	/*
	 * Return to caller with appropriate return code.
	 */
	return rc;
}

/**************************************************************************************
*
* Function:		ClearFile()
*
* Description:	Clears ALL records from a file.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the file was cleared, any other it the error number.
*
**************************************************************************************/

static int 
ClearFile ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             rc, connections_number;
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	/*
	 * Simply call Reality to clear the file.
	 */
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	rc = RfcClear (optr->FileHandle);	/* Clear the file of all records */
	/*
	 * Return to caller with appropriate return code.
	 */
	return rc;
}

/**************************************************************************************
*
* Function:		SelectFile()
*
* Description:	Begin the select of all items in a file.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the file was selected okay, any other it the error number.
*
**************************************************************************************/

static int 
SelectFile ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             rc, connections_number;
	RLC_LIST        ListHandle;
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	/*
	 * Call Reality to select the file.
	 */
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	if ((rc = RlcMakeList (optr->FileHandle, &ListHandle)) != 0)
	{
		return rc;
	}
	/*
	 * Return the list handle to the caller.
	 */
	outbuff.operand2 = (long) ListHandle;
	/*
	 * Return to the caller with successfule completion code.
	 */
	return 0;
}

/**************************************************************************************
*
* Function:		SelectEndFile()
*
* Description:	Terminate the selection of a file.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the file was successfully un-selected, any other it the error number.
*
**************************************************************************************/

static int 
SelectEndFile ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             connections_number;
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	/*
	 * Simply call Reality to clear the file.
	 */
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	/*
	 * Call REALITY to terminate the selection.
	 */
	return RlcCloseList ((RLC_LIST) inbuff.operand4);
}

/**************************************************************************************
*
* Function:		Readnext()
*
* Description:	Read the next record key from the selection list.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the next record key was read in okay, any other is an
*				error , specifically ENOENT means no more records.
*
**************************************************************************************/

static int 
Readnext ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             rc, connections_number;
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	/*
	 * Simply call Reality to clear the file.
	 */
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	rc = RlcNext ((RLC_LIST) inbuff.operand4, outbuff.buff, sizeof (outbuff.buff), &outbuff.operand1);
	/*
	 * Convert the return code.
	 */
	switch (rc)
	{
	case RLE_ENDOFLIST:
		/*
		 * No more records.
		 */
		rc = ENOENT;
		break;
	}
	/*
	 * Return to caller with appropriate return code.
	 */
	return rc;
}

/**************************************************************************************
*
* Function:		StatFile()
*
* Description:	Return textual information about the file.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	None -- we call the messages ourselves.
*
**************************************************************************************/

static void 
StatFile ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             i1, connections_number, rlen;
	char            buffer[REALITY_STAT_BUFF_SIZE], *p1;
	char            user[4096], account[4096];
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		SendBaseMessage (EINVAL);
		return;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		SendBaseMessage (EINVAL);
		return;
	}
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	/*
	 * Set up the default information for the user name
	 * and the account name. We should also strip out the
	 * passwords used if necessary.
	 */
	strcpy (user, (cptr->user == NULL ? "UNKNOWN" : cptr->user));
	if ((p1 = strchr (user, ',')) != NULL)
	{
		*p1 = '\0';
	}
	strcpy (account, (cptr->account == NULL ? "UNKNOWN" : cptr->account));
	if ((p1 = strchr (account, ',')) != NULL)
	{
		*p1 = '\0';
	}
	/*
	 * Make up some textual information.
	 */
	p1 = buffer;
	p1 += sprintf (p1, "File Type:      REALITY-X\n");
	p1 += sprintf (p1, "Database name:  %s\n", cptr->dbase);
	p1 += sprintf (p1, "User name:      %s\n", user);
	p1 += sprintf (p1, "Account Name:   %s\n", account);
	p1 += sprintf (p1, "File name:      %s\n", optr->filename);
	/*
	 * Send a base message with an appropriate return code.
	 */
	*p1++ = '\0';		/* Make the message NULL terminated */
	outbuff.operand1 = (p1 - buffer);	/* The length of the generated message */
	SendBaseMessage (0);	/* Send the base message */
	/*
	 * We have sent the base message.
	 * Now to send the textual data separately.
	 */
	rlen = outbuff.operand1;	/* Number of bytes still to send */
	p1 = buffer;		/* Address of next data byte to send */
	while (rlen > 0)
	{
		i1 = min (REALITY_BLOCK_SIZE, rlen);	/* Number of bytes to transmit this time */
		SendBytes (p1, i1);	/* Send this block of data */
		rlen -= i1;	/* Adjust number of bytes left to send */
		p1 += i1;	/* Adjust the pointer to the next data to send */
	}
}

/**************************************************************************************
*
* Function:		Lock()
*
* Description:	Perform locks an unlocks.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the operation worked, anything else is an error.
*
**************************************************************************************/

static int 
Lock ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             rc, connections_number;
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	/*
	 * Decide what sort of operation to do.
	 * It can be one of :
	 *   Lock record and wait.
	 *   Lock record non-blocking.
	 *   Unlock record.
	 *   Unlock all records on the file.
	 */
	if (inbuff.operand2 & JEDI_RECORD_UNLOCKRECORD)
	{
		/*
		 * Unlock a single record or ALL records.
		 */
		if (inbuff.operand1 < 0)
		{
			/*
			 * Unlock ALL records.
			 */
			rc = RfcUnlockAll (optr->FileHandle);
		}
		else
		{
			/*
			 * Unlock an individual record.
			 */
			rc = RfcUnlock (optr->FileHandle, inbuff.buff, inbuff.operand1);
		}
	}
	else if (inbuff.operand2 & JEDI_RECORD_LOCKRECORD)
	{
		/*
		 * Lock a record and wait.
		 */
		rc = EINVAL;	/* The REALITY driver doesn't support this */
	}
	else if (inbuff.operand2 & JEDI_RECORD_LOCKRECORD_NOWAIT)
	{
		/*
		 * Lock a record , non blocking.
		 */
		rc = EINVAL;	/* The REALITY driver doesn't support this */
	}
	else
	{
		rc = EINVAL;
	}
	/*
	 * Return to caller with appropriate return code.
	 */
	return rc;
}

/**************************************************************************************
*
* Function:		DeleteFile()
*
* Description:	Delete an existing file.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the operation worked, anything else is an error.
*
**************************************************************************************/

static int 
DeleteFile ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             rc, connections_number;
	char            filename[4096];
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	SET_ACCOUNT (cptr);	/* Ensure we are set to the correct account */
	/*
	 * The way we attempt to delete files in jBASE is slightly
	 * different. In jBASE we delete each section separately so typically
	 * we delete the DATA section then the DICT section. On Reality it is
	 * different, if we delete DATA section then both DICT and DATA
	 * will be deleted.
	 */
	if (memcmp (optr->filename, "DICT ", 5) == 0)
	{
		/*
		 * Simply Deleting the DICT section.
		 */
		strcpy (filename, optr->filename);
	}
	else if (strchr (optr->filename, ',') != NULL)
	{
		/*
		 * Simply Deleting a section in the file.
		 */
		strcpy (filename, optr->filename);
	}
	else
	{
		/*
		 * Copying a single DATA section.
		 */
		sprintf (filename, "%s,%s", optr->filename, optr->filename);
	}
	/*
	 * Now to attempt to delete the file.
	 */
	rc = RfcDeleteFile (filename);
	return rc;
}

/**************************************************************************************
*
* Function:		CreateFile()
*
* Description:	Create a file or a stub.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the operation worked, anything else is an error.
*
**************************************************************************************/

static int 
CreateFile ()
{
	int             i1, rc, rlen, connect_number, flags, exists_flag;
	char           *buffer, *p1, *filename, *args, argsbuffer[256];
	/*
	 * Get the number of bytes we will need for the rest
	 * of the definition and allocate that space.
	 */
	i1 = rlen = inbuff.operand1;
	if ((buffer = p1 = jmalloc (i1 + 10)) == NULL)
	{
		return errno;
	}
	/*
	 * Read in those bytes.
	 */
	while (rlen > 0)
	{
		i1 = min (REALITY_BLOCK_SIZE, rlen);	/* Number of bytes to recieve this time */
		RecieveBytes (p1, i1);	/* Recieve this block of data */
		rlen -= i1;	/* Adjust number of bytes left to recieve */
		p1 += i1;	/* Adjust the pointer to the next data to recieve */
	}
	*p1++ = '\0';		/* Ensure it is null terminated */
	/*
	 * Parse the passed buffer and attempt to open the connection.
	 */
	if ((rc = ParseFileConnect (buffer, &filename, &args, &connect_number, &exists_flag)) != 0)
	{
		jfree (buffer);
		return rc;
	}
	/*
	 * Handle the DICT file name.
	 */
	if (memcmp (filename, "DICT ", 5) == 0)
	{
		flags = RFC_OPT_DICT;
		filename += 5;
	}
	else
	{
		flags = 0;
	}
	/*
	 * Create the 'args'. This should be in the format 'nn,nn' such
	 * as 23,1. As jBASE will default the separation to 1, we will
	 * attempt to do the same.
	 */
	strncpy (argsbuffer, (args == NULL ? "" : args), sizeof (argsbuffer) - 4);
	argsbuffer[sizeof (argsbuffer) - 4] = '\0';
	if ((p1 = strchr (argsbuffer, ',')) == NULL)
	{
		/*
		 * Default the file separation to 1.
		 */
		strcat (argsbuffer, ",1");
	}
	/*
	 * Now to create the file.
	 */
	if ((rc = RfcCreateFile (RFC_DEFAULT_FILE, flags | RFC_OPT_MOD_SEP, filename, argsbuffer)) != 0)
	{
		/*
		 * If the create failed because the file existed, and the user had 'exists=true' on
		 * the command line, then we should allow this to work okay
		 * and just create a stub file.
		 */
		if (!exists_flag || rc != RFE_SECTEXISTS)
		{
			jfree (buffer);
			return rc;
		}
		rc = 0;
	}
	/*
	 * All went okay. Release any space used and return
	 * to the caller with a successful completion code.
	 */
	jfree (buffer);
	return rc;
}

/**************************************************************************************
*
* Function:		GetFileName()
*
* Description:	Get the name of an existing file.
*
* Parameters:	None -- all taken from inbuff .
*
* Return Value:	0 shows the operation worked, anything else is an error.
*
**************************************************************************************/

static int 
GetFileName ()
{
	struct openfiles *optr;
	struct connections *cptr;
	int             connections_number;
	/*
	 * Check the validity of the file handle.
	 */
	if (inbuff.handle < 0 || inbuff.handle >= openfiles_qty)
	{
		return EINVAL;
	}
	optr = openfiles_list + inbuff.handle;
	if (!optr->inuse)
	{
		return EINVAL;
	}
	connections_number = optr->connections;		/* The connection number */
	cptr = connections_list + connections_number;	/* Point to the connections handle */
	/*
	 * Return to the caller the name of the file.
	 */
	strncpy (outbuff.buff, optr->filename, sizeof (outbuff.buff) - 1);
	outbuff.buff[sizeof (outbuff.buff) - 1] = '\0';
	return 0;
}

/*********************************************************************************
*
* Function:		jcompare(string1 , string2 )
*
* Description:	Compare 2 strings to be equal. This also allows for
*				both strings being NULL pointers.
*
* Parameters:	string1:	First string to compare
*				string2:	Second string to compare.
*
* Return Value:	1 shows the strings compare, 0 means they differ.
*
*********************************************************************************/

static int 
jcompare (char *string1, char *string2)
{
	if (string1 == NULL)
	{
		if (string2 == NULL)
		{
			return 1;
		}
		if (*string2 == '\0')
		{
			/*
			 * We assume comparing NULL with "" is the same.
			 */
			return 1;
		}
		return 0;
	}
	if (string2 == NULL)
	{
		if (*string1 == '\0')
		{
			return 1;
		}
		return 0;
	}
	return (strcmp (string1, string2) == 0 ? 1 : 0);
}

/*********************************************************************************
*
* Function:		jstrdup(target , source )
*
* Description:	Duplicate a string, also allows NULL to be duplicated.
*
* Parameters:	target:		Where to store the result.
*				source:		The source to duplicate , can be NULL.
*
* Return Value:	0 shows it worked, 1 is an error (with errno set up)
*
********************************************************************************/

static int 
jstrdup (char **target, char *source)
{
	/*
	 * Check for NULL as the source.
	 */
	if (source == NULL)
	{
		*target = NULL;
		return 0;
	}
	if ((*target = strdup (source)) == NULL)
	{
		return 1;
	}
	return 0;
}

/***************************************************************************
*
* Function:		SetSignalHandlers()
*
* Description:	Set up all the signal handlers.
*
* Parameters:	None.
*
* Return Value:	None.
*
*******************************************************************************/

static void 
SetSignalHandlers ()
{
	if ((char *) sigset (SIGINT, sighandler) == (char *) SIG_IGN)
	{
		sigset (SIGINT, SIG_IGN);	/* Reinstate ignore disposition */
	}
	if ((char *) sigset (SIGQUIT, sighandler) == (char *) SIG_IGN)
	{
		sigset (SIGQUIT, SIG_IGN);	/* Reinstate ignore disposition */
	}
	if ((char *) sigset (SIGHUP, sighandler) == (char *) SIG_IGN)
	{
		sigset (SIGHUP, SIG_IGN);	/* Reinstate ignore disposition */
	}
	sigset (SIGUSR1, sighandler);
	sigset (SIGILL, sighandler);
	sigset (SIGSEGV, sighandler);
	sigset (SIGBUS, sighandler);
	sigset (SIGTERM, sighandler);
	sigset (SIGPIPE, sighandler);
	sigset (SIGFPE, sighandler);
	sigset (SIGABRT, sighandler);
	sigset (SIGALRM, sighandler);
}

/*******************************************************************************
*
* Function:		sighandler(signo)
*
* Description:	This is the signal handler.
*
* Parameters:	signo:	The signal.
*
* Return Value:	None.
*
*******************************************************************************/

static void 
sighandler (int signo)
{
	/*
	 * Choose what signals to ignore.
	 */
	switch (signo)
	{
	case SIGALRM:
	case SIGUSR1:
	case SIGUSR2:
	case SIGCHLD:
		SetSignalHandlers ();
		return;
	}
	fprintf (stderr, "%s: Error , signal %d seen\n", ProgramName, signo);
	Wrapup (2);
}

/******************************************************************************
*
* Function:		ParseFileConnect(inptr , filename , args , connect_number , exists_flag )
*
* Description:	Parse an input string that specifies a file on a remote
*				REALITY system and attempt to make a connection to that database.
*
* Parameters:	inptr:			A NULL terminated string that defines the input file.
*								THIS PARAMETER GETS CORRUPTE DURING A DESTRUCTIVE PARSE.
*				filename:		We return here the name of the file we parse. As it is a pointer
*								to somewhere in 'inptr' you will need to jmalloc or strdup this
*								to ensure it is preserved.
*				args:			We return here the optional arguments in the 'args=xxx' parameter.
*								Same limitations as for 'filename'.
*				connect_number:	We return here the connection number.
*				exists_flag:	We return 1 if 'exists=true|yes' is seen.
*
* Return Value:	0 shows all worked okay, anything else is an error.
*
**********************************************************************************/

static int 
ParseFileConnect (char *inptr, char **filename, char **args,
		  int *connect_number, int *exists_flag)
{
	struct connections *cptr, *unused, *found;
	char           *dbase, *user, *userpassword, *account, *accountpassword;
	char           *p1, *nextarg;
	int             i1, rc;
	/*
	 * Initialise the parameters we might find.
	 */
	dbase = user = userpassword = account = accountpassword = *filename = *args = NULL;
	*exists_flag = 0;
	/*
	 * Parse the command parameters.
	 */
	p1 = inptr;
	while (*p1 != '\0')
	{
		/*
		 * Skip white space.
		 */
		while (*p1 == '\t' || *p1 == '\0')
		{
			p1++;
		}
		if (*p1 == '\0')
		{
			break;
		}
		/*
		 * Now to parse again until we find some delimiter.
		 */
		nextarg = p1;
		while (*p1 != '\0' && *p1 != ' ' && *p1 != '\t')
		{
			p1++;
		}
		if (*p1 != '\0')
		{
			*p1++ = '\0';
		}
		/*
		 * See what the argument means.
		 */
		if (JediBaseUmemcmp (nextarg, "dbase=", 6) == 0)
		{
			dbase = nextarg + 6;
		}
		else if (JediBaseUmemcmp (nextarg, "user=", 5) == 0)
		{
			user = nextarg + 5;
		}
		else if (JediBaseUmemcmp (nextarg, "userpassword=", 13) == 0)
		{
			userpassword = nextarg + 13;
		}
		else if (JediBaseUmemcmp (nextarg, "account=", 8) == 0)
		{
			account = nextarg + 8;
		}
		else if (JediBaseUmemcmp (nextarg, "accountpassword=", 16) == 0)
		{
			accountpassword = nextarg + 16;
		}
		else if (JediBaseUmemcmp (nextarg, "args=", 5) == 0)
		{
			*args = nextarg + 5;
		}
		else if (JediBaseUmemcmp (nextarg, "exists=", 7) == 0)
		{
			if (JediBaseUmemcmp (nextarg + 7, "YES", -1) == 0 ||
			    JediBaseUmemcmp (nextarg + 7, "TRUE", -1) == 0)
			{
				*exists_flag = 1;
			}
		}
		else if (JediBaseUmemcmp (nextarg, "filename=", 9) == 0)
		{
			*filename = nextarg + 9;
			/*
			 * See if this is specifying a DICT section.
			 */
			if (strcmp (*filename, "DICT") == 0 && *p1 != '\0')
			{
				*(p1 - 1) = ' ';	/* Restore the white space */
				/*
				 * Skip past the white space until we come to the real file name .
				 */
				while (*p1 == ' ' || *p1 == '\t')
				{
					p1++;
				}
				/*
				 * Now to find and terminate (as before) the real file name.
				 */
				while (*p1 != '\0' && *p1 != ' ' && *p1 != '\t')
				{
					p1++;
				}
				if (*p1 != '\0')
				{
					*p1++ = '\0';
				}
			}
		}
		/*
		 * else { Don't worry about things we don't understand.}
		 */
	}
	/*
	 * For those things that didn't have an entry, see if there is
	 * an environment variable that will do.
	 */
	if (dbase == NULL)
	{
		if ((dbase = getenv ("jREALITY_DBASE")) == NULL)
		{
			dbase = getenv ("REALDBASE");
		}
	}
	if (user == NULL)
	{
		if ((user = getenv ("jREALITY_USER")) == NULL)
		{
			user = getenv ("REALUSER");
		}
	}
	if (userpassword == NULL)
	{
		userpassword = getenv ("jREALITY_USERPASSWORD");
	}
	if (account == NULL)
	{
		if ((account = getenv ("jREALITY_ACCOUNT")) == NULL)
		{
			account = getenv ("REALACC");
		}
	}
	if (accountpassword == NULL)
	{
		accountpassword = getenv ("jREALITY_ACCOUNTPASSWORD");
	}
	/*
	 * The only mandatory item is the file name.
	 */
	if (*filename == NULL || **filename == '\0')
	{
		return EINVAL;
	}
	/*
	 * Have a look around the list of connections to see if we are
	 * already connected.
	 */
	for (i1 = 0, cptr = connections_list, unused = NULL, found = NULL;
	     i1 < connections_qty;
	     i1++, cptr++)
	{
		if (cptr->inuse)
		{
			if (found == NULL &&
			    jcompare (cptr->dbase, dbase) &&
			    jcompare (cptr->user, user) &&
			    jcompare (cptr->userpassword, userpassword) &&
			    jcompare (cptr->account, account) &&
			  jcompare (cptr->accountpassword, accountpassword))
			{
				found = cptr;
			}
		}
		else
		{
			unused = cptr;
		}
	}
	/*
	 * If we haven't found a similar connection, then open a new connection.
	 */
	if (found == NULL)
	{
		if ((rc = RfcConnect (dbase, user, userpassword, account, accountpassword)) != 0)
		{
			return rc;
		}
		/*
		 * Found an empty space to put it in.
		 */
		if (unused == NULL)
		{
			if ((connections_list = (connections_list == NULL ? jmalloc (sizeof (struct connections)) : jrealloc (connections_list, sizeof (struct connections) * (connections_qty + 1)))) == NULL)
			{
				return errno;
			}
			unused = connections_list + connections_qty++;
		}
		/*
		 * Fill in the empty space.
		 */
		unused->inuse = 1;
		RfcGetAccount (&unused->AccountHandle);
		if (jstrdup (&unused->dbase, dbase) ||
		    jstrdup (&unused->user, user) ||
		    jstrdup (&unused->userpassword, userpassword) ||
		    jstrdup (&unused->account, account) ||
		    jstrdup (&unused->accountpassword, accountpassword))
		{
			return errno;
		}
		found = unused;
		if (RealityDebug)
		{
			fprintf (stderr, "%s: Connected to database %s , connection number %d\n", ProgramName, dbase, found - connections_list);
		}
	}
	/*
	 * Make this the current account.
	 */
	RfcSetAccount (found->AccountHandle);
	CurrentAccountHandle = found->AccountHandle;
	/*
	 * All seemed to go okay. Return the connection number
	 * to the caller and return with a successful completion code.
	 */
	*connect_number = found - connections_list;
	return 0;
}

/*******************************************************************************
*
* Function:		Usage()
*
* Description:	Display to stderr the usage of this program.
*
* Parameters:	None.
*
* Return Value:	None.
*
*******************************************************************************/

static void 
Usage ()
{
	fprintf (stderr, "Called as :\n");
	fprintf (stderr, "\t%s: {-h?}\n", ProgramName);
	fprintf (stderr, "Where :\n");
	fprintf (stderr, "\t-h or ?\tDisplay this help screen\n");
	fprintf (stderr, "%s Currently at revision %s\n", ProgramName, version);
}
