#include <jsystem.h>
#include <dirent.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <macros.h>
#include <sys/signal.h>
#include <sys/ipc.h>
#include <sys/msg.h>

/*
 * NOTE:	When listing or editing this source you will find a device of 132 columns
 *          with tabs set to 4 will produce the most pleasing result. Trying to edit this
 *          on an 80 column screen will prove much more strained.
 *
 * This is a demonstation use of an jBASE exteral device driver.
 * It is fairly simple - it performs it's reads and writes simply
 *    to the DEMO_DIRECTORY_NAME directory.
 * It is called on a one-to-one basis in that each active OPEN statement
 *    will have a copy of this process active, and it's own message queue.
 *    In real life, if this were to be used significantly, you would
 *    use the 'openid' and 'NewMessageId' fields in the 'JediMessageAcknowledge'
 *    return structure to show future calls are re-directed to another message
 *    queue - the resident server process - see later notes.
 *
 * This program is called as :
 *    jdevDEMO -i MessageKey -p ProcessId -o OptionalArgs
 *
 * Note the '-i MessageKey', '-p ProcessId' and '-o OptionalArgs' options are 
 *    created at run-time by the OPEN function. The actual description of this
 *    external file is thus :
 *       JBC__EDIjdevDEMO
 *
 * The -i option shows the initial message Id to use for communications.
 * The -o option shows the optional arguments passed during the OPEN
 *    statement, and are simply ignored in this program.
 * The -p option shows the process id of who called us.
 *
 * For your own device driver, any command line arguments you may decide
 *   upon please use upper case arguments except -J . This means you will
 *   not clash with any current or futute uses of command line
 *   arguments built in to jBASE.
 *
 * As this is a demo program, any unknown arguments are simply ignored.
 *
 * Notice that when the program exits, there is nothing to do. There are no
 *    message queues to delete, comms links to close, files and record locks
 *    to delete and so on. If there was, you would need to write some common
 *    wrapup code to handle all cases when the program exits. The program
 *    can exit for the following reasons :
 *       (a)     The CLOSE request is seen, and the main loop in the main() function
 *               is broken with the 'ExitFlag' variable.
 *       (b)     We get an error during the msgrcv() function.
 *       (c)     We see a signal and the SignalHandler() function is entered.
 */

static	int		MaxQLength = 0 ;											/* Maximum size of message queue */
static	char	*ProgramName ;
static	int		ExitCode = 0 ;

extern	int		optind , opterr ;
extern	char	*optarg ;

#define	PERROR(msg)	{fprintf(stderr,"%s: ",ProgramName);fflush(stderr);perror(msg);}

#define	DEMO_DIRECTORY_NAME		"/tmp" 

#define		MAX_SELECT		10											/* Maximum number of SELECT's on same file. */

static	DIR		*SelectDir[MAX_SELECT] ;

#define		CHAR_AM		0xfe 

void	SetSignals();
int		SignalHandler();

main(argc,argv)
int argc;
char *argv[];
{

	int		c1;
	char	*OptionalArg = NULL ;
	int		MsgQid = (-1);												/* Message queue id passed back by msgget() */
	int		OpenId = 0;													/* The unique ID we returned to caller. */
	int		minread , processid , ReturnCode , BytesRead , BytesWrite ;
	struct	JediMessageSend	*RecieveBuffer = NULL;						/* Data space for receiving messages */
	struct	JediMessageRecieve	*SendBuffer = NULL;						/* Data space for sending messages */
	struct	JediMessageAcknowledge Acknowledge ;						/* Structure to return the original OPEN request */

	umask(0);															/* Allow full read/write priveleges */

	SetSignals();														/* Set signal handlers */

	ProgramName = jstrdup(argv[0]);										/* Save for other functions to use globally */

	setpgrp();															/* Allow us to run as background - no logout or kill */

	srand48(time('\0'));												/* Initialise random number generator */

/*
 * Parse the command line arguments ourselves.
 */
	while((c1=getopt(argc , argv , "i:o:p:")) != (-1)) {
		switch(c1) {
		case 'i':														/* Message Id */
			MsgQid = strtol(optarg,(char*)0,0);							/* Convert string in hex, decimal or octal to integer */
			break;
		case 'o':														/* Optional command line argument */
			OptionalArg = jstrdup(optarg);
			break;
		case 'p':														/* Process Id of who called us */
			processid = strtol(optarg,(char*)0,0);						/* Convert string in hex, decimal or octal to integer */
			break;
		default:														/* Ignore anything we do not understand */
			break;
		}
	}

	if (MsgQid == (-1)) {												/* No message key number specified */
		fprintf(stderr,"%s: No message key specified\n",ProgramName);
		exit(2);
	}

/*
 * Get the stats about the queue.
 */
	MaxQLength = JediDemonQsize();
/*
 * Allocate space , equivalent to the maximum message length for both recieving
 *   messages and a separate buffer for sending messages.
 */
	if ((RecieveBuffer=(struct JediMessageSend *)jmalloc(MaxQLength + sizeof(long))) == NULL || (SendBuffer=(struct JediMessageRecieve *)jmalloc(MaxQLength + sizeof(long))) == NULL) {
		PERROR("jmalloc");
		exit(5);
	}

/*
 * Send an acknowledgement to the caller.
 *
 * NOTE:
 * This demonstration program will always acknowledge and accept
 *   this open request. It does not have to open comms links or check
 *   if the resource is already in use. If you have code that needs to be
 *   done at this stage, and it fails, the following code segment would
 *   cause the OPEN request to fail gracefully :
 *
 *          Acknowledge.type = 1 ;
 *          Acknowledge.ReturnCode = EIO ;		
 *          sprintf(Acknowledge.ErrorMessage,"Cannot open comms link");
 *          msgsnd(MsgQid,&Acknowledge,sizeof(Acknowledge)-sizeof(long),0);
 *          exit(1);
 *
 * Also, this demonstation program exist on a one-for-one basis. This means
 *   that for every OPEN that calls this device driver, a copy of this
 *   process exists while the OPEN is in effect (although as Unix uses shared
 *   text space, this is not so bad). This works well if only a few copies 
 *   are usually ever needed, but if you expect to have dozens of consecutive
 *   users, or want to manage a shared resource in a single program, or stream
 *   resources through a number of controlled threads, then you can let
 *   this OPEN succeed, but re-direct any future file i/o requests
 *   regarding this OPEN using the 'NewMessageId' and 'openid' members of
 *   the JediMessageAcknowledge structure. We will fill in the 'NewMessageId'
 *   member to show the client what message queue to use in future (the client
 *   recognises this and deletes any existsing private message queue). We will
 *   fill in 'openid' with a unique number, so that the new server process can
 *   uniquely identify future calls. The 'openid' is passed in every message
 *   between the client and the server. The following code segment shows
 *   how this would be achieved :
 *
 *          if ( ! IsServerActiveAlready() ) {
 *                   system("/usr/bin/IBMServer &");
 *          }
 *          Acknowledge.type = 1 ;
 *          Acknowledge.openid = GetOpenIdFromIBMServer(processid);
 *          Acknowledge.NewMessageId = GetMessageIdFromIBMServer(processid);
 *          Acknowledge.ReturnCode = 0 ;		
 *          strcpy(Acknowledge.ErrorMessage,"");
 *          msgsnd(MsgQid,&Acknowledge,sizeof(Acknowledge)-sizeof(long),0);
 *          exit(1);
 *  
 */
	Acknowledge.type = 1;												/* This field is not used */
	Acknowledge.ReturnCode = 0;											/* Show the open went okay */
	Acknowledge.openid = OpenId = lrand48();							/* Unique ID for the OPEN */
	Acknowledge.NewMessageId = MsgQid ;									/* Use the original message queue id */
	strcpy(Acknowledge.ErrorMessage,"");								/* Return a NULL error message to caller */
	if (msgsnd(MsgQid , &Acknowledge , sizeof(Acknowledge) - sizeof(long) , 0) < 0) {
		PERROR("msgsnd");
		exit(6);
	}
	
/*
 * Now go in a loop waiting for messages.
 * Note that the structure for the buffer we recieve into, 'RecieveBuffer'
 *   is mapped out to 'struct JediMessageSend', as this is the message
 *   that is sent by the client, and therefore recieved by us. This is
 *   conversely true for the send buffer, 'SendBuffer', which is mapped
 *   to 'struct JediMessageRecieve' as we must make it up for the client
 *   to recieve.
 */
	while(!ExitCode) {
		errno = 0 ;
		minread = sizeof(struct JediMessageSend)-sizeof(long);			/* Split to circumvent Interactive 386 'C' compiler error */
		if ((BytesRead=msgrcv(MsgQid,RecieveBuffer,MaxQLength,0x7fffffff,0)) < minread) {
			/*
			 * The msgrcv() has failed. This is pretty fatal stuff. Remember,
			 * even if a jBASE program is killed with an error, a signal (other
			 * than -9), or exits through the debugger etc., then it should
			 * still clean up and call the Close function. The main way
			 * this can fail at this point is because the message queue has
			 * been deleted. This will usually be by the operator using
			 * 'ipcrm -q nnn' or the client program has been killed with
			 * a SIGKILL signal, and therefore cannot clean up properly. In this
			 * case, the message queue will have been deleted by Unix doing
			 * it's own clean up on a terminated program with a message queue
			 * created with the IPC_PRIVATE flag.
			 *
			 * So, whatever the reason for us aborting, we should clean up at
			 * this stage, such as removing all record locks, closing down
			 * any comms links, deleting shared memory and so on. For this
			 * demonstration program, there is none of this to do, so we will
			 * just display an error message and exit.
			 */
			PERROR("msgrcv()");
			exit(errno);
		}
		BytesWrite = 0;
/*
 * We have recieved a message from somebody. Hopefully, it is from the same process
 *    that initiated us. We can check this by using the 'openid' field of all sent
 *    messages, and make sure that it is the same as the one we generated earlier
 *    and returned to the caller during acknowledgement.
 */
		if (RecieveBuffer->openid != OpenId) {
			ReturnCode = EACCES ;										/* Note same user, so give an error code */
		} else {
			switch(RecieveBuffer->command) {							/* See what command was requested */
			case	JEDI_MESSAGE_COMMAND_CLOSE:							/* Close the device/file */
				ReturnCode = DemoClose(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_READ:							/* Read from the device */
				ReturnCode = DemoRead(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_WRITE:							/* Write to the device */
				ReturnCode = DemoWrite(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_DELETE:						/* Delete record from the device */
				ReturnCode = DemoDelete(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_CLEAR:							/* Clear all records from the device. */
				ReturnCode = DemoClear(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_SELECT:						/* Start the selection of all records from device */
				ReturnCode = DemoSelect(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_SELECT_END:					/* End the selection of all records from device */
				ReturnCode = DemoSelectEnd(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_READNEXT:						/* Read the records from device after a SELECT */
				ReturnCode = DemoReadnext(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_SYNC:							/* Synchronise databases, etc. */
				ReturnCode = DemoSync(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_LOCK:							/* Various lock functions, such as release all */
				ReturnCode = DemoLock(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			case	JEDI_MESSAGE_COMMAND_IOCTL:							/* Various IOCTL operations */
				ReturnCode = DemoIoctl(RecieveBuffer,BytesRead,SendBuffer,&BytesWrite);
				break;
			default:													/* Invalid command seen */
				fprintf(stderr,"%s: Invalid command number %d seen\n",ProgramName,RecieveBuffer->command);
				ReturnCode = EINVAL;
			}
		}

/*
 * Now to write the reply to the caller. 
 * The following variables have already been created :
 *    (i)    The 'SendBuffer' structure, containing many union's, will have been filled in
 *           by the relevent function and is entirely functionally dependent.
 *    (ii)   The value in 'ReturnCode' will be 0 if all worked okay, or any other +ve
 *           value to denote an error condition.
 *    (iii)  The value in BytesWrite shows the number of bytes to write after
 *           the address given by ((char*)SendBuffer) + sizeof(JediMessageRecieve)
 */
		SendBuffer->type = RecieveBuffer->returntype;
		SendBuffer->returncode = ReturnCode ;
		msgsnd(MsgQid,SendBuffer,BytesWrite+sizeof(struct JediMessageRecieve)-sizeof(long),0);
	}

/*
 * The CLOSE request has been seen.
 * Some device drivers may wish to remain active after the CLOSE request. We have
 *   been operating on a one-to-one basis, so we will simply exit.
 */

	exit(0);

}

/***********************************************************
*
* Function:		DemoClose(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Close the device.
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means all worked okay, any other is an error.
*
* Notes:		None.
*
*******************************************************************/

int		DemoClose(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		BytesWrite ;
{


/*
 * For this test program, there is nothing much to do.
 * Setting 'ExitCode' causes the main() function to exit the main loop
 *    after writing back a reply. There are no files to close or
 *    communication links to close. As we have used the message queue
 *    provided by the client, there is no message queue to delete - the
 *    client will close it's own message queue once it has recieved the
 *    reply from this function.
 *
 * NOTE:
 *    If there is a record locking scheme implemented, then this Close
 *    request must unlock ALL locks held by this device. A seperate call
 *    to the Lock function will not be made prior to the Close, so you
 *    must release all outstanding locks before the close.
 *
 *    As this is a simple demonstration program that does not keep
 *    any record locks, there is nothing to do about it in this example code.
 */
	ExitCode = 1 ;

	return(0);

}

/***********************************************************
*
* Function:		DemoWrite(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Write a record. For this demo, we simply write the
*				data to the DEMO_DIRECTORY_NAME directory. Note that as the length of
*				the data to write may exceed the maximum length of a 
*				message, there may be multiple calls to this function
*               for a single write.
*
*               As we are writing to a Unix directory, we will assume in 
*               this simple demo program that new-lines are the Unix
*               equivalent to attribute marks, so we will replace one with
*               the other. If the client wanted to write in binary, avoiding
*               any translations, we would use the IOCTL mechanism to let
*               the client inform us that future writes are in binary (this
*               mechanism is already implemented on the Unix directory
*               device driver, which is an integral part of jBASE).
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means all worked okay, any other is an error.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoWrite(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

	static	char	*RecordKey = NULL , *Record = NULL;
	char	*DataPtr , *ptr1 , *ptr2 ;
	unsigned char	*u1 , *u2 ;
	char	PathName[2048] ;
	int		i1 , fd ;
	static	int		RecordLen ;

/*
 * The following fields in the 'Recieve' pointer will be used.
 *    (i)     WRITE_RECORD_PREVIOUS:          Number of bytes sent for this record in previous
 *                                            messages. If this value is 0, then it is the first
 *                                            message for this write request.
 *    (ii)    WRITE_RECORD_LEN:               Total size of the record to write.
 *    (ii)    WRITE_RECORD_THIS_MSG:          Number of bytes for the data record contained
 *                                            in this message only. If (WRITE_RECORD_PREVIOUS+WRITE_RECORD_THIS_MSG)
 *                                            equals WRITE_RECORD_LEN, then this is the last message for the write.
 *    (iv)    WRITE_RECORD_KEY_LEN:           If the value here is >= 0, then the data that follows the
 *                                            JediMessageSend structure, as pointed to by 'Recieve', 
 *                                            is the record key to use during the write.
 *    (v)     WRITE_FLAGS:                    Allows locks and so on to be taken. See the JEDI_RECORD_xxx
 *                                            definitions in jedi.h. Ignored for this demo driver.
 *    (vi)    WRITE_FIELD_NUMBER:             Used to support WRITEV and READV operations. We do not support
 *                                            these operations in this device driver. The function for the WRITE
 *                                            request knows what drivers support field reads/writes, and so will
 *                                            use other mechanisms if it sees a READV or WRITEV request. In other
 *                                            words, this field is ignored.
 *    (vii)   WRITE_CONVERT_FLAGS:            Allows us to perform (or not perform) conversions between new-lines
 *                                            and attribute marks. See the JEDI_CONVERT_xxx definitions in jedi.h .
 *                                            For this simple device driver, we will ALWAYS perform the conversions,
 *                                            and so this field is ignored.
 */

/*
 * Extract the record key if necessary.
 */
	DataPtr = ((char*)Recieve)+sizeof(struct JediMessageSend);			/* The address of any data to write out */
	if (Recieve->WRITE_RECORD_KEY_LEN >= 0) {							/* See if user has specified a record key */
		RecordLen = Recieve->WRITE_RECORD_LEN;							/* Extract length of data for this record */
		if (RecordKey != NULL) jfree(RecordKey);							/* Release space used by any existing definition */
		if ((RecordKey=(char*)jcalloc(1,Recieve->WRITE_RECORD_KEY_LEN+1))==NULL) return errno;	/* Allocate space for record key */
		memcpy(RecordKey,DataPtr,Recieve->WRITE_RECORD_KEY_LEN);		/* Create a null terminated string for record key */
		DataPtr += Recieve->WRITE_RECORD_KEY_LEN;						/* This is the start of any following data */
	}

/*
 * Create the buffer big enough for entire record if necessary.
 */
	if (Recieve->WRITE_RECORD_PREVIOUS == 0 && Record != NULL) {
		/*
		 * Time to clean up previous buffers.
		 */
		jfree(Record);
		Record = NULL;
	}
	if (Record == NULL) {
		if ((Record = jmalloc(RecordLen+2)) == NULL) return errno;		/* Allocate enough space for entire record. */
	}

/*
 * Copy just THIS segment of data to the buffer.
 */
	memcpy(Record+Recieve->WRITE_RECORD_PREVIOUS , DataPtr , Recieve->WRITE_RECORD_THIS_MSG);

/*
 * See if this is the LAST message to complete the write.
 */
	if (Recieve->WRITE_RECORD_PREVIOUS + Recieve->WRITE_RECORD_THIS_MSG < RecordLen) return 0;

/*
 * We are the last message.
 * Convert any attribute mark characters to new-line.
 */
	i1 = RecordLen  ; u1 = (unsigned char *) Record;					/* Initialise to start of buffer */
	while(i1 > 0 && (u2 = memchr(u1 , CHAR_AM , i1)) != NULL) {
		*u2 = '\n';														/* Replace attribute mark with new-line */
		u2++;															/* Point one past the previous delimiter */
		i1 -= (u2-u1);													/* Adjust number of bytes remaining */
		u1 = u2;
	}
/*
 * Write a trailing new-line to the buffer. We have previously allocated enough space
 *   for this.
 */
	*(Record + RecordLen) = '\n' ;

/*
 * Finally, write out the record to the DEMO_DIRECTORY_NAME directory.
 */
	sprintf(PathName , "%s/%s",DEMO_DIRECTORY_NAME,RecordKey);			/* Make up a full unix key */
	if ((fd = open(PathName , O_WRONLY|O_TRUNC|O_CREAT , 0666)) < 0) {	/* Try to open the file */
		return errno;
	}
	write(fd , Record , RecordLen+1);									/* Write out, don't check return code to be lazy */
	close(fd);															/* Close the file after use */

/*
 * Finished writing the data. Clean up the space used.
 */
	jfree(RecordKey);
	jfree(Record);
	RecordKey = Record = NULL;

	return(0);

}

/***********************************************************
*
* Function:		DemoRead(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Read a record. For this demo, we simply read the
*				data from the DEMO_DIRECTORY_NAME directory. Note that as the length of
*				the data to read may exceed the maximum length of a 
*				message, there may be multiple calls to this function
*               for a single read.
*
*               As we are writing to a Unix directory, we will assume in 
*               this simple demo program that new-lines are the Unix
*               equivalent to attribute marks, so we will replace one with
*               the other. If the client wanted to read in binary, avoiding
*               any translations, we would use the IOCTL mechanism to let
*               the client inform us that future reads are in binary (this
*               mechanism is already implemented on the Unix directory
*               device driver, which is an integral part of jBASE).
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means all worked okay, any other is an error.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoRead(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

	static	char	*Record , *RecordKey ;
	static	int		RecordLen , BytesWritten , BytesLeft ;
	char	*ptr1 , PathName[2048] ;
	int		fd , saveerrno , i1 ;
	unsigned char *u1 , *u2 ;
	struct	stat StatInfo ;

/*
 * The following fields in the 'Recieve' pointer will be used.
 *    (i)     READ_RECORD_KEY_LEN:            If the value here is >= 0, then the data that follows the
 *                                            JediMessageSend structure, as pointed to by 'Recieve', 
 *                                            is the record key to use during the read.
 *    (ii)    READ_FLAGS:                     Allows locks and so on to be taken. See the JEDI_RECORD_xxx
 *                                            definitions in jedi.h. Ignored for this demo driver.
 *    (iii)   READ_FIELD_NUMBER:              Used to support WRITEV and READV operations. We do not support
 *                                            these operations in this device driver. The function for the READ
 *                                            request knows what drivers support field reads/writes, and so will
 *                                            use other mechanisms if it sees a READV or WRITEV request. In other
 *                                            words, this field is ignored.
 *    (iv)    READ_CONVERT_FLAGS:             Allows us to perform (or not perform) conversions between new-lines
 *                                            and attribute marks. See the JEDI_CONVERT_xxx definitions in jedi.h .
 *                                            For this simple device driver, we will ALWAYS perform the conversions,
 *                                            and so this field is ignored.
 *    (v)     READ_BLOCK_NUMBER:              Shows the relative message number used to satsify this single READ
 *                                            request, so if it is 0 then it is the first request for the data.
 */

/*
 * Extract the record key if necessary.
 */
	if (Recieve->READ_RECORD_KEY_LEN >= 0) {							/* See if user has specified a record key */
		ptr1 = ((char*)Recieve)+sizeof(struct JediMessageSend);			/* The address of any data to read */
		if (RecordKey != NULL) jfree(RecordKey);							/* Release space used by any existing definition */
		if ((RecordKey=(char*)jcalloc(1,Recieve->READ_RECORD_KEY_LEN+1))==NULL) return errno;	/* Allocate space for record key */
		memcpy(RecordKey,ptr1,Recieve->READ_RECORD_KEY_LEN);			/* Create a null terminated string for record key */
	}

/*
 * If this is the first time round, then read in the data all in one go after
 *    opening the file.
 */
	if (Recieve->READ_BLOCK_NUMBER == 0) {
		if (Record != NULL) {
				/* Time to clean up previous buffers. */
			jfree(Record);
			Record = NULL;
		}
		/*
		 * Open the requested file, get information about the file (in particular it's size),
		 *   and allocate enough space for entire record.
		 */
		sprintf(PathName,"%s/%s",DEMO_DIRECTORY_NAME,RecordKey);		/* Make up the path name to open */
		if ((fd = open(PathName , O_RDONLY , 0)) < 0) return errno;		/* Make the initial open request */
		if (fstat(fd , &StatInfo) != 0) {								/* Get information about the file */
			saveerrno = errno; close(fd) ; return saveerrno ;
		}
		RecordLen = StatInfo.st_size ;									/* Get size of record */
		if ((Record = jmalloc(RecordLen)) == NULL) {						/* Allocate enough space for the record */
			saveerrno = errno; close(fd) ; return saveerrno ;
		}
		/*
		 * Read in the entire record in one go.
		 */
		if (read(fd , Record , RecordLen) != RecordLen) {
			saveerrno = errno; close(fd) ; return (saveerrno == 0 ? EIO : saveerrno) ;
		}
		close(fd);														/* Close file, don't need it anymore */

		/*
		 * If the last character in the record is a new-line character, we can
		 *    assume it is a text file , rather than a binary file and we should convert
		 *    all new-lines to attribute marks, stripping off the final new-line character.
		 * This is crude but works more time. If you want more control, then use the IOCTL mechanism
		 *    described elsewhere in this source code.
		 */
		if (*(Record+RecordLen-1) == '\n') {
			RecordLen--;												/* Ignore the trailing new-line */
			i1 = RecordLen  ; u1 = (unsigned char *) Record;			/* Initialise to start of buffer */
			while(i1 > 0 && (u2 = memchr(u1 , '\n' , i1)) != NULL) {
				*u2 = CHAR_AM;											/* Replace new-line with attribute mark */
				u2++;													/* Point one past the previous delimiter */
				i1 -= (u2-u1);											/* Adjust number of bytes remaining */
				u1 = u2;
			}
		}
		BytesWritten = 0;												/* Initialise number of bytes written so far */
		BytesLeft = RecordLen ;											/* Initialise number of bytes still to write back */
	}

/*
 * Make up the data in the return buffer.
 */
	i1 = min((MaxQLength - sizeof(struct JediMessageRecieve)) , BytesLeft);	/* Number of bytes to transfer this go */
	Send->READ_RECORD_SIZE = RecordLen ;								/* Always give the record length */
	Send->READ_RECORD_PREVIOUS = BytesWritten ;							/* Bytes written in previous calls */
	Send->READ_RECORD_THIS_MSG = i1;									/* Number of bytes supplied this time round */
	memcpy(((char*)Send)+sizeof(struct JediMessageRecieve) , Record + BytesWritten , i1);	/* Copy supplied data */
	BytesWritten += i1 , BytesLeft -= i1;								/* Adjust statistics about what has gone so far */
	*BytesWrite = i1 ;													/* Adjust size of message to send back */

	if (BytesLeft <= 0) {
		/*
		 * We are the last message for this record. Clean up the space used.
		 */
		jfree(RecordKey);
		jfree(Record);
		RecordKey = Record = NULL;
	}

	return(0);

}

/***********************************************************
*
* Function:		DemoDelete(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Delete a record. For this demo, we simply delete the
*				record from the DEMO_DIRECTORY_NAME directory. 
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means all worked okay, any other is an error.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoDelete(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

	char	PathName[2048] ;
	int		i1 ;

/*
 * The following fields in the 'Recieve' pointer will be used.
 *    (i)     DELETE_RECORD_KEY_LEN:          This is the length of record key to delete.
 *    (ii)    DELETE_FLAGS:                   Allows locks and so on to be taken. See the JEDI_RECORD_xxx
 *                                            definitions in jedi.h. Ignored for this demo driver.
 */

/*
 * Extract and create the record key.
 */
	i1 = sprintf(PathName,"%s/",DEMO_DIRECTORY_NAME);				/* Make up the original directory path */
	memcpy(&PathName[i1] , ((char*)Recieve)+sizeof(struct JediMessageSend) , Recieve->DELETE_RECORD_KEY_LEN);
	PathName[i1 + Recieve->DELETE_RECORD_KEY_LEN] = NULL;			/* Make it a null terminated string */
	return (unlink(PathName) == 0 ? 0 : errno);

}

/***********************************************************
*
* Function:		DemoClear(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Delete ALL records from a file. For this demo, we simply clear the
*				records from the DEMO_DIRECTORY_NAME directory. 
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 is USUALLY returned, denoting success. We do not delete
*				any sub-directories, nor do we delete empty directories, nor
*				worry about any files we cannot delete due to lack of
*				permissions. A non-zero error may otherwise be returned,
*				denoting an error.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoClear(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

/*
 * The following fields in the 'Recieve' pointer will be used.
 *    None.
 */

	DIR		*DirPtr;
	struct	dirent	*Next;
	char	PathName[PATH_MAX*2];

	if ((DirPtr = opendir(DEMO_DIRECTORY_NAME)) == NULL) return(errno);	/* Perform the SELECT on the directory */

	while ((Next = readdir(DirPtr)) != NULL) {			/* Get next directory entry name */
		strcpy(PathName,DEMO_DIRECTORY_NAME);			/* Copy the directory name */
		strcat(PathName,"/");							/* Add a delimiter */
		strcat(PathName,Next->d_name);					/* Copy the file name within the directory */
		unlink(PathName);								/* Delete the file, ignore any errors */
	}

	closedir(DirPtr);									/* Close the unix directory */

	return (0);

}

/***********************************************************
*
* Function:		DemoSelect(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Start the selection of all records from the device file.
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means the select worked okay, any other means a failure.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoSelect(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

	int		SelectEntry ;

/*
 * The following fields in the 'Recieve' pointer will be used.
 *    None.
 */

/*
 * Try to find a jfree entry in the SelectDir[] matrix.
 */
	for (SelectEntry = 0 ; SelectEntry < MAX_SELECT ; SelectEntry++) {
		if (SelectDir[SelectEntry] == NULL) break;
	}
	if (SelectEntry == MAX_SELECT) {
		/* Too many SELECT statements performed on this one server. */
		return EMFILE;
	}
/*
 * Now to initiate a select on this directory.
 */
	if ((SelectDir[SelectEntry] = opendir(DEMO_DIRECTORY_NAME)) == NULL) return(errno);	/* Perform the SELECT on the directory */

/*
 * Return to caller the select entry we used. This is passed back to
 *   us for future reference in future READNEXT and SELECTEND calls.
 */
	sprintf(Send->SELECT_ID,"%d",SelectEntry);

	return 0;
}

/***********************************************************
*
* Function:		DemoSelectEnd(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	End the selection of all records from the device file.
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means the select worked okay, any other means a failure.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoSelectEnd(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

	int		SelectEntry ;
/*
 * The following fields in the 'Recieve' pointer will be used.
 *    (i)     SELECTEND_ID:                   This is the id of the entry in the select table , 
 *                                            as returned by the SELECT command to uniquely identify
 *                                            multiple selects on the same file.
 */

/*
 * Get the select entry number, from 0 to (MAX_SELECT-1).
 */
	if (sscanf(Recieve->SELECTEND_ID,"%d",&SelectEntry) != 1) return EINVAL ;
	if (SelectEntry < 0 || SelectEntry >= MAX_SELECT) return EINVAL ;

	closedir(SelectDir[SelectEntry]);									/* Close the unix directory */
	SelectDir[SelectEntry] = NULL;

	return 0 ;

}

/***********************************************************
*
* Function:		DemoReadnext(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Read the next record key following a previous SELECT.
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means the select worked okay, any other means a failure.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoReadnext(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

	struct	dirent	*Next;
	int		SelectEntry , i1 ;
/*
 * The following fields in the 'Recieve' pointer will be used.
 *    (i)     READNEXT_ID:                    This is the id of the entry in the select table , 
 *                                            as returned by the SELECT command to uniquely identify
 *                                            multiple selects on the same file.
 */

/*
 * Get the select entry number, from 0 to (MAX_SELECT-1).
 */
	if (sscanf(Recieve->READNEXT_ID,"%d",&SelectEntry) != 1) return EINVAL ;
	if (SelectEntry < 0 || SelectEntry >= MAX_SELECT) return EINVAL ;

	if ((Next = readdir(SelectDir[SelectEntry])) == NULL) {
		return ENOENT;				/* Get next directory entry name */
	}

	i1 = strlen(Next->d_name);											/* Get length of name */
	*BytesWrite = i1 ;													/* Show main() how much EXTRA to write */
	Send->READNEXT_RECORD_KEY_LEN = i1 ;								/* Show client how long the record key is */
	memcpy(((char*)Send)+sizeof(struct JediMessageRecieve) , Next->d_name , i1);	/* Copy the actual name */

	return 0 ; 
}

/***********************************************************
*
* Function:		DemoSync(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Synchronise the database by flushing any cache, writing
*				out scratch buffers and so on. For this demonstration
*				program, it has no effect.
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means the select worked okay, any other means a failure.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoSync(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

/*
 * The following fields in the 'Recieve' pointer will be used.
 *    None.
 */

/*
 * In this demonstration program there is nothing to do. For more complex
 *   applications, you may want to flush out buffers, or call the
 *   Unix 'flush(filedescriptor)' function and so on.
 */
	return 0 ;
}

/***********************************************************
*
* Function:		DemoLock(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Various mechanisms for unlocking records. For this
*				simple demonstration program that does not support any
*				locking mechanism, nothing happens. However, we will
*				document in the code, later on, what the meaning of all
*               the passed variables are and how you should respond yourself
*				if your own driver DOES support locking.
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means the select worked okay, any other means a failure.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoLock(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

	int		RecordKeyLen , Flags ;
	char	RecordKey[260];												/* Note the maximum record key size is 255 bytes */

/*
 * The following fields in the 'Recieve' pointer will be used.
 *    (i)     LOCK_FLAGS:                     Combinations of the JEDI_RECORD_xxx bits to show if we
 *                                            are to lock and return , lock and wait or unlock
 *                                            a record or file.
 *    (ii)    LOCK_RECORD_KEY_LEN:            Length of record key that follows, and shows what
 *                                            record is to be locked (with or without waiting),
 *                                            or to be unlocked. This can be a value of -1 on an
 *                                            unlock request, showing that ALL locks on the file
 *                                            are to be released.
 */

	RecordKeyLen = Recieve->LOCK_RECORD_KEY_LEN;						/* Extract record key length */
	Flags = Recieve->LOCK_FLAGS ;										/* Extract the flags */

	if (RecordKeyLen >= 0) {
		memcpy(RecordKey , ((char*)Recieve)+sizeof(struct JediMessageSend) , RecordKeyLen);
		RecordKey[RecordKeyLen] = NULL;
	} else {
		RecordKey[0] = NULL;
	}

/*
 * Now to test what action to perform.
 * For this demo program, it actually performs nothing - the code is
 *    there just to demonstrate the passed parameters.
 */
	if (Flags & JEDI_RECORD_LOCKRECORD) {
		/*
		 * Lock the record given in buffer 'RecordKey', and wait ad-infinitum 
		 *    until the program is killed, or the lock becomes jfree.
		 */
	} else if (Flags & JEDI_RECORD_LOCKRECORD_NOWAIT) {
		/*
		 * Lock the record given in buffer 'RecordKey'.
		 * If the record lock is busy, then return immediately with a 
		 *    return code of JEDI_ERRNO_LOCK_TAKEN.
		 * If the record lock CAN be taken, then take it and return 0.
		 */
	} else if (Flags & JEDI_RECORD_UNLOCKRECORD) {
		if (RecordKeyLen < 0) {
			/*
			 * Although this is an unlock request, no record key is provided.
			 * This means we want to release ALL locks on this file.
			 */
		} else {
			/*
			 * Unlock the record whose record key is in 'RecordKey'.
			 * Do NOT report any errors if the record is not locked
			 *    to start with, just NO-OP the command.
			 */
		}
	} else {
		/*
		 * This code point should theoretically never be reached, as 
		 *   it means the function was called with nothing to do.
		 */
	}

/*
 * For this demo program, always return 0 which signifies the lock/unlock
 *    was taken, or the unlock was taken but was not locked already. If
 *    a value of JEDI_ERRNO_LOCK_TAKEN is returned, this means the
 *    JEDI_RECORD_LOCKRECORD_NOWAIT bit was set and the record was already
 *    locked. Any other returned value is a serious error.
 */
	return 0;
}

/***********************************************************
*
* Function:		DemoIoctl(Recieve , BytesRead , Send , BytesWrite)
*
* Description:	Various commands can be passed here by the IOCTL
*				function in the jBASE language.
*
* Parameters:	Recieve:	Address of a JediMessageSend buffer, which
*							is the data sent to us from the client.
*				BytesRead:	Number of bytes read in the message. This
*							includes the JediMessageSend buffer and any data
*							that follows it, but not the first long value.
*				Send:		Address of a JediMessageRecieve buffer, where we
*							make up the return message. The 'type' and 
*							'returncode' fields are filled in by the main
*							function.
*				BytesWrite:	Address of an int value, where we return the
*							number of bytes we have used that exceed the
*							JediMessageRecieve structure. The default is
*							0 bytes.
*				
* Return Value:	0 means the select worked okay, any other means a failure.
*
* Notes:		Transaction boundaries , Transaction Logging and record locks
*				are not supported in this demonstration driver.
*
*******************************************************************/

int		DemoIoctl(Recieve , BytesRead , Send , BytesWrite)
struct	JediMessageSend	*Recieve ;
int		BytesRead ;
struct	JediMessageRecieve	*Send ;
int		*BytesWrite ;
{

	int		ReturnCode = 0 , i1 ;
	char	*ptr1 ;
/*
 * The following fields in the 'Recieve' pointer will be used.
 *    (i)     IOCTL_COMMAND:                  The actual command to perform. See the later description.
 *    (ii)    IOCTL_TEXTLEN:                  Number of bytes following Send structure to modify the command.
 */

/*
 * The way IOCTL works is like this:
 * In a jbc program, you may have the following code segment.
 *
 *      INCLUDE JBC.h
 *      OPEN "FILEX" TO DSCB ELSE STOP 201,"FILEX"
 *      Param = "ABCDEF"
 *      IF IOCTL(DSCB , JIOCTL_COMMAND_GETSTATUS , Param) THEN
 *          CRT "This IOCTL worked , Param = ":Param
 *      END ELSE
 *          CRT "The IOCTL failed" ; STOP
 *      END
 *
 * The first parameter to IOCTL() is the opened file descriptor. The second is the command
 *    you want the device driver to perform. There are a number of pre-defined commands
 *    defined, for use with jbc programs, in the supplied JBC.h header file, and matching
 *    #define statements are supplied with the jedi.h header file for the device drivers
 *    written in 'C' . The third parameter is a variable whose contents get passed to the
 *    device driver, and whose contents get updated and returned by the device driver. The
 *    meaning and usage of this third parameter depends entirely upon the command being
 *    requested by the second parameter.
 * You can use this mechanism to allow your application in jbc to communicate with your
 *    own device driver. The disadvantage is that this then becomes highly device dependent
 *    and the IOCTL() will only work with the device driver you have specifically coded it for.
 *    However, this is the reason for having IOCTL() - that is non-generic and device specific
 *    communications with a device driver.
 * If you write your own IOCTL code, then the second parameter passed to IOCTL() should be
 *    greater than 1000 to avoid conflicts with future internal uses of the IOCTL() function
 *    by the jBASE libraries.
 * In the IOCTL() function in a jbc program, the second parameter, the command to perform, can
 *    be found in the 'Recieve->IOCTL_COMMAND' integer. The third parameter, any optional
 *    qualifiers for the command, can be found in the text string following the 'Send'
 *    structure for a length of 'Recieve->TEXTLEN' bytes.
 *
 * There are a number of IOCTL commands built in to jBASE and are defined as JIOCTL_COMMAND_xxx
 *    equates in the JBC.h header file for jbc sources, and their equivalent as JEDI_IOCTL_xxx
 *    #define's in the jedi.h header file for 'C' sources. All of these can be safely ignored
 *    on this device driver, as they are pretty device specific.
 *
 * To enable a better understanding of how you would code your own IOCTL, we will show how
 *    this device driver can respond to the following code segment :
 *
 *          INCLUDE JBC.h
 *          EQUATE GET_HOST_NAME TO 1000 , GET_SERVER_PID TO 1001
 *          OPEN "FILENAME" TO DSCB ELSE STOP 201,"FILENAME"
 *          Msg = ""
 *          IF IOCTL(DSCB , GET_HOST_NAME , Msg) ELSE
 *              CRT "IOCTL FAILED" ; STOP
 *          END
 *          CRT "Name of host for file FILENAME = ":Msg
 *          Msg = "Process ID = "
 *          IF IOCTL(DSCB , GET_SERVER_PID , Msg) ELSE
 *              CRT "IOCTL FAILED" ; STOP
 *          END
 *          CRT Msg
 *
 *   The output from this program , using the code below, would typically be :
 *          Name of host for file FILENAME = DEMO.PROGRAM
 *          Process ID = 20971
 */
	switch(Recieve->IOCTL_COMMAND) {
	case 1000:															/* Get the host name */
		/*
		 * For this demo program, we will hard code the name of the
		 *    host to be 'DEMO.PROGRAM', so we will simply return this string.
		 */
		strcpy(((char*)Send)+sizeof(struct JediMessageRecieve) , "DEMO.PROGRAM");
		*BytesWrite = strlen("DEMO.PROGRAM");							/* Return to main() length of extra data to return */
		break;
	case 1001:															/* Get our pid */
		/*
		 * We will return a string showing the process ID of this process.
		 * By means of an exercise only, we will precede the process id with any
		 *   text string passed to us in the first instance.
		 */
		/*
		 * Work out how many bytes of the passed string we will use. This
		 *   depends upon the maximum length of a message (MaxQLength) , the size already
		 *   required for the return message header (sizeof(struct JediMessageRecieve)),
		 *   10 bytes reserved for the process id string, and the actual size of
		 *   the returned message.
		 */
		i1 = min((MaxQLength - sizeof(struct JediMessageRecieve))-10 , Recieve->IOCTL_TEXTLEN);
		ptr1 = ((char*)Send)+sizeof(struct JediMessageRecieve);			/* Target destination address */
		/*
		 * Now copy, to the target destination as pointed to by ptr1, the 
		 *   original string passed to us and the process id.
		 */
		memcpy(ptr1 , ((char*)Recieve)+sizeof(struct JediMessageSend) , i1);
		i1 += sprintf(ptr1 + i1 , "%d" , getpid());						/* Append process id to our return string */
		*BytesWrite = i1 ;												/* Return to main() optional byte count */
		break;
	default:															/* Unknown IOCTL */
		ReturnCode = EINVAL;
		break;
	}

	Send->IOCTL_RETURNLEN = *BytesWrite ;								/* A safeguard for integrity */

	return ReturnCode ; 

}

/****************************************************************************
*
* Function:		SetSignals()
*
* Description:	Set up signal handlers.
*
* Parameters:	None.
*
* Return Value:	None.
*
* Notes:		None.
*
*******************************************************************************/

void	SetSignals()
{

	signal(SIGINT,SignalHandler);
	signal(SIGUSR1,SignalHandler);
	signal(SIGILL,SignalHandler);
	signal(SIGSEGV,SignalHandler);
	signal(SIGBUS,SignalHandler);
	signal(SIGTERM,SignalHandler);
	signal(SIGPIPE,SignalHandler);
	signal(SIGQUIT,SignalHandler);
	signal(SIGHUP,SignalHandler);
	signal(SIGFPE,SignalHandler);
	signal(SIGABRT,SignalHandler);
	signal(SIGALRM,SignalHandler);
}

SignalHandler(sig)
int		sig;
{

	char	Message[512];
	/*
	 * Restore the signal handler.
	 */
	signal(sig , SignalHandler );

/*
 * We have seen a signal of some sort.
 * In ANY event, we will exit and not carry on any more.
 * As this is a demonstration program, and the caller established a private
 *    message queue, there is nothing more to do. There are no queues to delete,
 *    no shared memory to delete, no comms links to close, no files or file
 *    locks to worry about -- in fact nothing.
 */
	
	exit(1);

}
