/*
 * NOTE: This source code is best viewed on a terminal with a width of 132
 *        characters, and tab stops set to 4.
 *
 * This is an example source to show how a normal 'C' program can
 * use the jedi API for file access. It uses most jedi calls to
 * copy all records from one file to another.
 *
 * The documentation 'External Interfaces for jBASE' describes the
 * mechanism more fully. 
 *
 * This source code would normally be compiled with something like this :
 *     % cc jediCExample.c -o CExample -I$JBCRELEASEDIR/include	\
 *				$JBCRELEASEDIR/lib/libjedi.so -lld
 * Or it could be compiled under Unix (but not AIX) as :
 *     % jbc jediCExample.c -o CExample
 * On NT 
 * 	  > cl /Zi -o CExample jediCExample.c /I%JBCRELEASEDIR%\include libjbase.lib
 *
 * It is then called as :
 *     Example {-cdo} {InputFileName {OutputFileName}}
 *
 * Option:
 *     -c       Clear the OutputFileName before starting any copies.
 *     -d       Delete the records from the InputFileName after copying the record.
 *     -o       Overwrite the record in the OutputFileName if it already exists.
 *
 * This example was compiled and tested under UnixWare version 1.1 and
 * AIX version 3.2. For other Unix versions you may need to amend some
 * of the Unix definitions slightly.
 *
 * This source code is for example purposes only. No responsibility can
 * be assumed for the correct working of this source.
 *
 */

 /*
  * Include files supplied with jBASE.
  */
#include <jsystem.h>

/*
 * Include files supplied with Unix.
 */

#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#undef	abs
#include <malloc.h>

#ifndef WIN32
#include <dirent.h>
#include <macros.h>
#endif

extern	int		optind ;


main(argc , argv)
int		argc ;
char	*argv[] ;
{

	int		c1 ;
	char	*p1 ;
	JediFileDescriptor	*Input , *Output ;
	struct	JediSelectPtr	*InputSelect ;
	char	RecordKey[512] , *RecordKeyPtr ;
	int		RecordKeyLenPtr ;
	int		CopyCount , DoWrite ;
	char	*ErrorMessage , ErrorMessageBuffer[256] ;
	char	*ReadBufferPtr , ReadBuffer[4096] ;
	int		ReadBufferLenPtr ;
	char	IoctlRtn[1024];
	int		IoctlRtnPtr ;
	int		ClearfileFlag = 0;				/* Set to 1 if -c option to clear target file */
	int		OverwriteFlag = 0 ;				/* Set to 1 if -o option to overwrite records */
	int		DeleteRecordFlag = 0 ;			/* Set to 1 if -d option to delete record from source after copying */
	char	InputFileName[1024],OutputFileName[1024] ;
	/*
	 * Parse the command line options.
	 */
	while((c1=getopt(argc,argv,"cdo"))!= (-1)) {
		switch(c1) {
		case 'c':
			/*
			 * -c option to clear target file.
			 */
			ClearfileFlag = 1 ;
			break;
		case 'd':
			/*
			 * -d option to delete record from source after copying it.
			 */
			DeleteRecordFlag = 1 ;
			break;
		case 'o':
			/*
			 * -o option to overwrite existing records.
			 */
			OverwriteFlag = 1 ;
			break;
		default:
			break;
		}
	}
	/*
	 * Get the name of the input file.
	 * If not specified on the command line, then prompt for it
	 * and read in from stdin.
	 */
	if (optind < argc) {
		strncpy(InputFileName,argv[optind++],sizeof(InputFileName)-1);
	} else {
		printf("Enter name of input file  : "); fflush(stdin);
		gets(InputFileName);
	}
	if (!strlen(InputFileName))
	{
		fprintf(stderr,"No input file name specified\n");
		exit(1);
	}
	/*
	 * Get the name of the output file.
	 * If not specified on the command line, then prompt for it
	 * and read in from stdin.
	 */
	if (optind < argc) {
		strncpy(OutputFileName,argv[optind++],sizeof(OutputFileName)-1);
	} else {
		printf("Enter name of output file : "); fflush(stdin);
		gets(OutputFileName);
	}
	if (!strlen(OutputFileName))
	{
		fprintf(stderr,"No output file name specified\n");
		exit(1);
	}
	/*
	 * Now to open both the files.
	 */
	if ((errno = JediOpen(&Input , InputFileName , NULL , 0 )) != 0)
	{
		perror(InputFileName);
		exit(2);
	}
	if ((errno = JediOpen(&Output , OutputFileName , NULL , 0 )) != 0)
	{
		perror(OutputFileName);
		JediClose(Input , 0);										/* Close the opened file */
		exit(2);
	}
	/*
	 * If the -c option was seen on the command line, the user
	 * wants to clear the output file at the start.
	 */
	if (ClearfileFlag)
	{
		printf("Clearing all records from file %s ...",OutputFileName);
		fflush(stdout);
		JediClearFile(Output) ;
		printf("\n");
		fflush(stdout);
	}
	/*
	 * Call the select function to select all record keys in the input file.
	 */
	if ((errno = JediSelect(Input , &InputSelect )) != 0)
	{
		perror(InputFileName);
		JediClose(Input , 0);										/* Close the opened file */
		JediClose(Output , 0);										/* Close the opened file */
		exit(2);
	}
	/*
	 * Now go in a loop reading each record from the input file and
	 * writing to the output file.
	 */
	for(CopyCount = 0 ;;)
	{
		/*
		 * Get the next record key.
		 */
		RecordKeyPtr = RecordKey ;
		RecordKeyLenPtr = sizeof(RecordKey);
		if ((errno = JediReadnext(Input , InputSelect , &RecordKeyPtr , &RecordKeyLenPtr)) != 0)
		{
			perror(InputFileName);
			JediSelectEnd(Input , InputSelect);							/* Terminate the selection process. */
			JediClose(Input , 0);										/* Close the opened file */
			JediClose(Output , 0);										/* Close the opened file */
			exit(3);
		}
		/*
		 * Exit this loop if all records have been selected.
		 */
		if (RecordKeyLenPtr < 0) 
		{
			/*
			 * There are now no more record keys from the select list
			 * generated in variable InputSelect. We will call JediSelectEnd() to terminate
			 * processing of this select list and to release all resources it may have
			 * allocated. 
			 */
			JediSelectEnd(Input , InputSelect);
			break;
		}
		/*
		 * See if the record already exists on the target file.
		 * If it does, then we do not perform the copy. This action
		 * can be suppressed with the -o option.
		 */
		DoWrite = 1;
		ErrorMessage = NULL ;
		if (!OverwriteFlag)
		{
			/*
			 * Test if the record is already there.
			 * We will use an IOCTL function to decide this, as this
			 * will be quicker than doing a READ. However, if the database
			 * driver doesn't support this in the IOCTL, then we will HAVE
			 * to do a READ to decide.
			 */
			IoctlRtnPtr = sizeof(IoctlRtn)-1 ;
			switch((errno = JediIOCTL(
							Output , 						/* File descriptor for the IOCTL command */
							JEDI_IOCTL_FINDRECORD ,			/* The actual command to perform */
							RecordKeyPtr , 					/* Start address of input data to IOCTL */
							RecordKeyLenPtr , 				/* Length of input data to IOCTL */
							IoctlRtn , 						/* Start address of returned data from IOCTL */
							&IoctlRtnPtr)))		 			/* Length of address of returned data */
			{
			case ENOENT:
				/*
				 * The database driver has accepted our IOCTL request.
				 * The record does not exist.
				 */
				break;
			case 0:
				/*
				 * The database driver has accepted our IOCTL request.
				 * The record already exists.
				 */
				ErrorMessage = "RecordExists" ;
				DoWrite = 0 ;
				break;
			default:
				/*
				 * One of two things here.
				 * Maybe the database driver does not support this IOCTL request,
				 * in which case -1 will have been returned. 
				 * Maybe there was a fatal error while trying to read the record.
				 * Whatever the reason, we will try to do a READ to see if
				 * the target record already exists.
				 */
				ReadBufferPtr = ReadBuffer ;
				ReadBufferLenPtr = sizeof(ReadBuffer);
				switch((errno = JediReadRecord(
							Output , 							/* File descriptor for the output file */
							0 , 								/* Flags for field number and locking */
							RecordKeyPtr , 						/* Record key */
							RecordKeyLenPtr ,					/* Length of record key */
							&ReadBufferPtr ,					/* Buffer to read into. */
							&ReadBufferLenPtr ,					/* Length of buffer to read into */
							0 ,									/* Field number to read in */
							jmalloc )))							/* The function to call to re-allocate space */
				{
				case 0:
					/*
					 * Record already exists.
					 */
					ErrorMessage = "RecordExists" ;
					DoWrite = 0 ;
					break;
				case ENOENT:
					/*
					 * Record does not exist.
					 */
					break;
				default:
					/*
					 * Some sort of fatal error.
					 */
					sprintf(ErrorMessageBuffer,"Error %d",errno);
					ErrorMessage = ErrorMessageBuffer ;
					DoWrite = 0 ;
					break;
				}
				/*
				 * Release any space allocated by JediReadRecord().
				 */
				if (ReadBufferPtr != ReadBuffer)
				{
					jfree(ReadBufferPtr);
				}
				break;
			}
		}
		if (DoWrite)
		{
			/*
			 * Now to actually do the write request.
			 * Before we do the write, we have to read the record in the first instance.
			 * When we call the READ function, we supplied the address of a static buffer.
			 */
			ReadBufferPtr = ReadBuffer ;
			ReadBufferLenPtr = sizeof(ReadBuffer);
			switch((errno = JediReadRecord(
						Input , 							/* File descriptor for the input file */
						0 , 								/* Flags for field number and locking */
						RecordKeyPtr , 						/* Record key */
						RecordKeyLenPtr ,					/* Length of record key */
						&ReadBufferPtr ,					/* Buffer to read into. */
						&ReadBufferLenPtr ,					/* Length of buffer to read into */
						0 ,									/* Field number to read in */
						jmalloc)))							/* The function to call to re-allocate space */
			{
			case 0:
				/*
				 * Record was read in successfully.
				 * Now to write out the same record to the output file.
				 */
				switch((errno = JediWriteRecord(
						Output , 							/* File descriptor for the output file */
						0 , 								/* Flags for field number and locking */
						RecordKeyPtr , 						/* Record key */
						RecordKeyLenPtr ,					/* Length of record key */
						ReadBufferPtr ,						/* Data to write out */
						ReadBufferLenPtr ,					/* Length of buffer to write out */
						0)))								/* Field number to write out */
				{
				case 0:
					/*
					 * Write was successful.
					 */
					break;
				default:
					/*
					 * Fatal write error.
					 */
					sprintf(ErrorMessageBuffer,"Error %d",errno);
					ErrorMessage = ErrorMessageBuffer ;
					break;
				}
				break;
			case ENOENT:
				/*
				 * Input record is not there. This is very unlikely to ever happen, 
				 * as the previous call to JediSelect has just said a few billiseconds
				 * ago that it existed. So, although this is very unlikely to happen, it
				 * could be somebody has just deleted the record between the JediSelect() call
				 * and this call to actually read the record.
				 */
				ErrorMessage = "No such record" ;
				break;
			default:
				sprintf(ErrorMessageBuffer,"Error %d",errno);
				ErrorMessage = ErrorMessageBuffer ;
				break;
			}
			/*
			 * Release any space allocated by JediReadRecord().
			 */
			if (ReadBufferPtr != ReadBuffer)
			{
				jfree(ReadBufferPtr);
			}
		}
		if (ErrorMessage == NULL)
		{
			/*
			 * The record was copied.
			 */
			CopyCount++;
			fwrite(RecordKeyPtr , 1 , RecordKeyLenPtr , stdout);
			printf("\n");
			/*
			 * If the -d option was seen, then delete the record from the source file.
			 */
			if (DeleteRecordFlag)
			{
				 switch((errno = JediDeleteRecord(
								Input ,
								0 ,
								RecordKeyPtr ,
								RecordKeyLenPtr )))
				{
				case 0:
					/*
					 * Record was delete successfully.
					 */
					break;
				case ENOENT:
					/*
					 * Record did not exist in the source to be deleted.
					 * This may seem strange as a page of code ago we read the
					 * record in. However, as we are not taking any record locks, and
					 * the record locks are advisory rather than mandatory, another
					 * user could have deleted the record in the time between reading
					 * it in and now trying to delete it.
					 */
					ErrorMessage = "No such record on source to be deleted" ;
					break;
				default:
					/*
					 * Indeterminate error while deleting the record.
					 */
					sprintf(ErrorMessageBuffer,"Error %d while deleting record from source",errno);
					ErrorMessage = ErrorMessageBuffer ;
					break;
				}
			}
		}
		if (ErrorMessage != NULL)
		{
			/*
			 * Some sort of error occured.
			 * Display the error message.
			 */
			fwrite(RecordKeyPtr , 1 , RecordKeyLenPtr , stderr);
			fprintf(stderr,": %s: NOT WRITTEN\n",ErrorMessage);
		}
		/*
		 * If the record key was too large for the buffer supplied
		 * (highly improbable but not impossible), then the database driver will
		 * have allocated its own memory space which we must now jfree.
		 */
		if (RecordKeyPtr != RecordKey)
		{
			jfree(RecordKeyPtr);
		}
	}
	printf("%d records copied\n",CopyCount);
	/*
	 * Close the input and output files.
	 * Most database drivers should be able to copy with a program
	 * terminating without actually closing a file -- this is because a
	 * program can terminate abnormally and you don't want to bring down
	 * your database just because of one rogue program.
	 *
	 * So, although not mandatory, we will close the files
	 * to correctly wrap them up.
	 */
	JediClose(Input , 0);										/* Close the opened file */
	JediClose(Output , 0);										/* Close the opened file */
	/*
	 * Exit the program.
	 */
	exit(0);
}
