/*
 * NOTE: This source code is best viewed on a terminal with a width of 132
 *        characters, and tab stops set to 4.
 *
 * This is an example source for shared object database drivers.
 * The documentation 'External Interfaces for jBASE' describes the
 * mechanism more fully. In brief, however, it allows a third party (be it
 * a software house, end user, VAR etc) to write their own database drivers, so
 * that jBASE applications can transparently access alternative databases.
 *
 * This source code would normally be compiled with something like this :
 *     % cc jediDExample.c -c -I$JBCRELEASEDIR/include
 * It can probably be better compiled using jcompile instead of cc. The jcompile command
 * adds options to the cc command on a per-platform basis to enable better
 * code portability. E.G.
 *     % jcompile jediDExample.c -o jediDExample.so
 *
 * Then, assuming that $HOME/lib was a component of the JBCOBJECTLIST 
 * environment variable, any jBASE application can open a file , and
 * perform file operations through this source assuming they have 
 * a file definition something like :
 *    JBC__SOB ExampleInit .d
 *
 * This database driver uses the normal Unix file system as a crude sort
 * of database. When an OPEN is executed, and our function ExampleOPEN
 * gets called, the directory it uses will be a concatenation of the file
 * name that contained the 'JBC__SOB ExampleInit .d' definition, plus the
 * ".d" string. This is only an example. You can change the source to access
 * any database you want. The creation of the file name in the host database
 * system is left entirely up to the writer of this database driver source.
 * As stated earlier, this example uses the concatenation of two string. You
 * could equally have chosen to just use the name in the file definition.
 * For example, the file definition could be changed to :
 *     JBC__SOB ExampleInit CUSTOMERS -v
 * and you could change this function to parse the "CUSTOMERS -v" string so
 * that you opened a C-ISAM file in /home/details/CUSTOMERS with user
 * verification of authority to do so.
 *
 * All the jBASE utilities such as jstat , jchmod, create-file etc. will only
 * work on the supported internal database types.
 *
 * This source code is for example purposes only. No responsibility can
 * be assumed for the correct working of this source.
 *
 */

/*
 * Include files supplied with jBASE.
 */
#include <jsystem.h>
#include <jedi.h>
/*
 * Include files supplied with Unix.
 */
#ifndef WIN32
#include <dirent.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <sys/stat.h>
#ifndef	__MVS__
#include <malloc.h>
#endif
#undef abs			/* Required for some platforms where abs() is duplicately defined */
#ifndef WIN32
#include <unistd.h>
#endif
#include <stdlib.h>


#ifndef WIN32
#define O_BINARY	0
#endif

#ifndef	PATH_MAX
#define	PATH_MAX	1024
#endif

#undef	min
#define	min(a,b) ((a) < (b) ? (a) : (b) )

struct	UnixSelect
{
	DIR		*SelectDIR;
};
struct	TypeExample
{
	char	* pathname ;
	int		pathnamelen ;
} ;

static	int		ExampleOPEN(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , char * HeaderInfo ,
					int HeaderInfoLen , int flags ) ;
static	int		ExampleCLOSE(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags) ;
static	int		ExampleSELECT(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor ,
					struct JediSelectPtr ** SelectPtr );
static	int		ExampleSELECTEND(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor ,
					struct JediSelectPtr * SelectPtr) ;
static	int		ExampleREADNEXT(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor ,
					struct JediSelectPtr * SelectPtr , char ** RecordKeyPtr , int * RecordLenPtr ) ;
static	int		ExampleREADRECORD(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags , char * RecordKey , 
					int RecordKeyLen , char ** BufferPtr , int * BufferLenPtr , int FieldNumber , void *(*mallocptr)(size_t size)) ;
static	int		ExampleWRITERECORD(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags ,
					char * RecordKey , int RecordKeyLen , char * BufferPtr , int BufferLen , int FieldNumber) ;
static	int		ExampleDELETERECORD(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags ,
					char * RecordKey , int RecordKeyLen ) ;
static	int		ExampleLOCK(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags , char * RecordKey ,
					int RecordKeyLen ) ;
static	int		ExampleIOCTL(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int SubCommand , void * IoctlAddr , 
					int IoctlLen , void * IoctlReturnAddr , int * IoctlReturnLen ) ;
static	int		ExampleCLEARFILE(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor ) ;
static	int		ExampleDELETEFILE(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int flags ) ;
static	int		ExampleSYNC(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor );
static	int		JediUnixMakeFullName(struct jBASEDataAreas * dp ,  JediFileDescriptor *FileDescriptor, char *RecordKey ,
					int  RecordKeyLen , char *NewAddr , int  NewLen );

#define		EXAMPLE_FILE_TYPE		"EXAMPLE"				/* Defines this database type */
#ifdef WIN32
#define		DIR_DELIM_CH			'\\'						/* Delimiter as a character */
#define		DIR_DELIM_STR			"\\"						/* Delimiter as a string */
#else
#define		DIR_DELIM_CH			'/'						/* Delimiter as a character */
#define		DIR_DELIM_STR			"/"						/* Delimiter as a string */
#endif

/***************************************************************************
*
* Function:		ExampleInit()
*
* Description:	Initialisation for the shared object example source.
*
*				This is the initialisation function that gets called when the
*				database driver is first referenced. It simply sets up a
*				structure defining all the references to file I/O functions,
*				and calls a jedi function to register us as a database driver.
*
* Parameters:	None:
*
* Return Value:	0 means all worked okay. Any other value is an error.
*
* Notes:		None.
*
***************************************************************************/

int		ExampleInit(struct jBASEDataAreas * dp)
{
	struct	JediDriverStruct	driver ;
	/*
	 * Initialise the members of the 'driver' structure with our device drivers.
	 */
	memset(&driver,0,sizeof(driver));
	/*
	 * The field 'FileTypeDescr' describes the type of file.
	 */
	driver.FileTypeDescr = EXAMPLE_FILE_TYPE ;							/* Description of file */
	driver.OpenCodePtr = ExampleOPEN ;									/* Function to perform the open file */
	driver.JediClose = ExampleCLOSE ;									/* Function to close the file */
	driver.JediSelect = ExampleSELECT ;									/* Function to perform a SELECT */
	driver.JediSelectEnd = ExampleSELECTEND ;							/* Function to call at end of SELECT */
	driver.JediReadnext = ExampleREADNEXT ;								/* Function to perform a READNEXT */
	driver.JediReadRecord = ExampleREADRECORD ;							/* Function to read a record */
	driver.JediWriteRecord = ExampleWRITERECORD ;						/* Function to write a record */
	driver.JediDeleteRecord = ExampleDELETERECORD ;						/* Function to delete a record */
	driver.JediLock = ExampleLOCK ;										/* Locking mechanism for the file */
	driver.JediIOCTL = ExampleIOCTL ;									/* Function to perform IOCTL */
	driver.JediClearFile = ExampleCLEARFILE ;							/* Function to clear all records from the file */
	driver.JediDeleteFile = ExampleDELETEFILE ;							/* Function to actually delete the file */
	driver.JediSync = ExampleSYNC ;										/* Function to synchonise file (flush cache) */
	driver.JediTransactionBegin = 
	driver.JediTransactionRollback = 
	driver.JediTransactionCommit = NULL ;								/* Show no internal transaction support */
	/*
	 * Add this to the list of supported database drivers.
	 * The JediBaseAddDriver will return the database driver number allocated 
	 * from 0 upwards, or -1 of an error occurs.
	 */
	return JediBaseAddDriver(dp,&driver);
}

/***********************************************************************************
*
* Function:		JediFileOpCreateEXAMPLE( filename , filetype , fileperm , fileflags , hashmethod , 
*					argc , argv , qualifiers , typebuffer , realpathname )
*
* Description:	This function will be called in an attempt to create a file
*				if the command line had 'type=example' in it. Note it is case
*				insensitive.
*
* Parameters:	filename:		The name of the file to create.
*				filetype:		The type of the file in mixed case , e.g. "exAmPle".
*				fileperm:		The permissions to create the file with.
*				fileflags:		Any number of the JEDI_FILE_xxx bits (found in jedi.h)
*								that are OR'ed together.
*				hashmethod:		If the command line had 'hashmethod=nnn' in it, this is
*								the numeric equivalent of nnn. If not, this value is 0.
*               argc:           The number of variable arguments passed.
*               argv:           The variable arguments.
*               qualifiers:     List of optional 'xx=yy' qualifiers.
*               typebuffer:     Address of buffer we update if we decide NOT to
*                               create a file of type HASH4.
*               realpathname:   We return here the real path name of the file created.
*
* Return Value:	We return 0 here if we created the file okay, or any other value
*				shows the create failed.
*
* Notes:		We should by this time have already checked the file does not exist,
*				but we will do this again !
*
*				The use of the 'qualifiers' allows you to add extra parameters to the
*				create-file syntax. It is up to the database driver to decide which
*				qualifiers to accept, and which to ignore. By convention, the
*				'perm=nnn' is generally the only one we MUST take any notice of, and
*				this has already been pre-parsed and passed in the 'fileperm' parameter.
*
*				The name of this function is significant -- it MUST be JediFileOpCreateEXAMPLE,
*				where EXAMPLE is the file type used in the create-file command, for example :
*					: create-file myfile type=example
*
************************************************************************************************/

int JediFileOpCreateEXAMPLE(struct jBASEDataAreas * dp , char * filename , char * filetype , int fileperm , int fileflags , 
						int hashmethod , int argc , char * argv[] , char ** qualifiers , char * typebuffer , char * realpathname )
{
	int		rc , fd ;
	struct	stat	statinfo ;
	char	dirname[4096] , header[1024] ;
	/*
	 * Adjust the permissions, so that we add the 'x' bit to
	 * each relevent part that has read permissions. Therefore, the user
	 * may create a file and the default permissions are 0644, but this is
	 * no good -- we need 0755.
	 */
	if (fileperm & 0600) fileperm |= 0100 ;			/* If user has read or write, add the execute bit */
	if (fileperm & 060) fileperm |= 010 ;			/* If group has read or write, add the execute bit */
	if (fileperm & 06) fileperm |= 01 ;				/* If other has read or write, add the execute bit */
	/*
	 * Return to caller the type of file created.
	 */
	strcpy(typebuffer,"EXAMPLE");
	/*
	 * Find out the actual path name to use. By default, this will be
	 * the same as 'filename', but if the user had entered 'createpath=openmax' then
	 * we want to create it in the directory where there is the most space.
	 *
	 * The call below to JediBaseModifyPath also sets up the return value
	 * 'realpathname' to be the path name used.
	 */
	JediBaseModifyPath(dp ,qualifiers , filename , 1024 , realpathname);
	/*
	 * Make the sub-directories in between the final path name and current path name.
	 */
	if ((errno=JediBaseCreateSubDirectories(realpathname,fileperm))) 
	{
		perror(realpathname);
		return errno ;
	}
	/*
	 * We will create two objects.
	 * (a) The actual file definition by the same as the path name
	 * (b) The directory associated with it. For this example only,
	 *     this will be the pathname + ".d"
	 * Make sure neither of these exist in the first instance.
	 */
	if (stat(realpathname , &statinfo) != 0)
	{
		if (errno != ENOENT)
		{
			perror(realpathname);
			return errno;
		}
	}
	sprintf(dirname , "%s.d",realpathname);
	if (stat(realpathname , &statinfo) != 0)
	{
		if (errno != ENOENT)
		{
			perror(realpathname);
			return errno;
		}
	}
	/*
	 * We have already made the sub-directories. Now for the last directory.
	 */
	errno = 0;
	if ((rc = mkdir(dirname,fileperm)) == 0)
	{
		if ((fd = open(realpathname,O_RDWR|O_CREAT|O_TRUNC|O_BINARY,fileperm)) < 0)
		{
			rc = errno ;
		}
		else
		{
			/*
			 * Write back a header for shared object drivers.
			 * The string 'JBC__SOB' is a sort of magic number to show
			 * jEDI it is a shared object driver.
			 * The string 'ExampleInit' is the name of the main entry function
			 * for the shared object driver, also contained in this source.
			 * The string '.d' is optional to database drivers, but shows in
			 * this case the directory name is filename + ".d".
			 */
			write(fd , header , sprintf(header,"JBC__SOB ExampleInit .d\n"));
			close(fd);
			rc = 0 ;
		}
	}
	else
	{
		rc = errno;
	}
	/*
	 * Pass back the correct return code to caller.
	 */
	return rc ;
}

/******************************************************************
*
* Function:		ExampleOPEN(FileDescriptor , HeaderInfo , HeaderInfoLen , flags )
*
* Description:	This is called when an OPEN request is being performed.
*
* Parameters:	FileDescriptor:	Address of a JediFileDescriptor variable type.
*				HeaderInfo:		The first 'HeaderInfoLen' bytes at the start of the file.
*				HeaderInfoLen:	The number of valid bytes pointed to by HeaderInfo.
*				flags:			General usage flags.
*
* Return Value:	0 means the file was opened okay. The other 'errno' values could be :
*					ENOENT	- We are not interested in this file type. However, there is
*								no reason not to try other file types.
*					EMFILE	- We have run out of file descriptors. Close a file and try again.
*					Others	- Fatal error -- do not try other file types.
*
*				Various elements of the FileDescriptor structure also need to be
*				filled in. See the 'External Interfaces for jBASE' documentation for
*				full details.
*
* Notes:		None.
*
***************************************************************/

static	int		ExampleOPEN(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , char * HeaderInfo ,
							int HeaderInfoLen , int flags )
{
	char				NewPathName[(PATH_MAX*2)+1] , *p1 , * pathname ;
	struct	stat		stat_info ;
	struct	TypeExample	* typeptr ;
	/*
	 * Create a new path name out of the components of
	 * the original Unix file that was opened, plus the operand
	 * in the file definition. For example, if the original Unix file
	 * opened was ./files/CUSTOMERS and looked like this :
	 *     JBC__SOB ExternalInit .d
	 * then we want to produce a string ./files/CUSTOMERS.d
	 * 
	 * As an added complication, we want to make sure that the path name
	 * becomes an absolute path name. This is because between the open
	 * and other file i/o commands, the user may, within the application,
	 * change their directory. This means that the file i/o operations
	 * within here would therefore be for the wrong Unix directory
	 * if the path was relative.
	 */
	pathname = JRunFileIOGetFullPath(dp , FileDescriptor->fdinfo);					/* Get the name of the file opened so far */
	if ((p1 = strchr(FileDescriptor->OptionalArgs,' ')) != NULL)
	{
		/*
		 * Get rid of any trailing spaces in the file name operand.
		 */
		*p1 = '\0';
	}
#ifdef WIN32
	if (*pathname == DIR_DELIM_CH || (*pathname+1 == ':' && *pathname+2 == DIR_DELIM_CH) ) 
#else
	if (*pathname != DIR_DELIM_CH) 
#endif
	{
		/*
		 * Path name for file descriptor is already an absolute path name.
		 */
		sprintf(NewPathName,"%s%s" , 
			pathname  ,												/* This is the absolute name of the file opened */
			FileDescriptor->OptionalArgs);							/* This is the suffix to create a directory name */
	}
	else
	{
		/*
		 * Create an absolute path name from a relative path name.
		 */
		sprintf(NewPathName,"%s%c%s%s" , 
			FileDescriptor->CurrentDir , 							/* This is the current directory we are in */
			DIR_DELIM_CH ,											/* Directory delimiter */
			pathname  ,												/* This is the relative name of the file opened */
			FileDescriptor->OptionalArgs);							/* This is the suffix to create a directory name */
	}
	/*
	 * Make sure this is a Unix directory.
	 */
	if (stat(NewPathName,&stat_info) != 0 || !S_ISDIR(stat_info.st_mode))
	{
		return ENOENT;
	}
	/*
	 * We want to save the name of this new directory.
	 * Luckily there is a pointer allocated in 'FileDescriptor' that
	 * can be used by database drivers as we please.
	 */
	if ((typeptr = calloc(1 , sizeof(struct TypeExample))) == NULL ||
		(typeptr->pathname = strdup( NewPathName)) == NULL)
	{
		return errno ;
	}
	FileDescriptor->TypePtr = typeptr ;
	typeptr->pathnamelen = strlen(typeptr->pathname);
	/* 
	 * Close the Unix file associated with this, if any.
	 */
	JRunFileIOClose(dp,&FileDescriptor->fdinfo , 0 );
	/* 
	 * Show we are using the internal locking scheme.
	 */
	FileDescriptor->Status |= JEDI_STATUS_INTERNAL_LOCKS ;	
	/*
	 * We will over-ride the default READ/WRITE flags so that we
	 * give them full write permissions -- the actual Unix read/write
	 * will work/fail anyway. Also, the file descriptor with the 'JBC__SOB' may
	 * only have certain permissions, but the actual directory will have different
	 * permissions, so let the operating system work it out.
	 */
	FileDescriptor->Status |= (JEDI_STATUS_READ|JEDI_STATUS_WRITE);	
	/*
	 * Show the type of file system is a user-defined one.
	 */
	FileDescriptor->FileType = JEDI_TYPE_USER ;
	/*
	 * The FileFlags member can be set to any number of the JEDI_FILE_xxx bits.
	 * These suppress operations such as transaction logging. In our case,
	 * we want all operations, so we set the flag to 0. (This is the default, 
	 * but performed here for clarity).
	 */
	FileDescriptor->FileFlags = 0 ;
	/*
	 * Return with a good completion code.
	 */
	return 0 ;
}

/*******************************************************************
*
* Function:		ExampleCLOSE(FileDescriptor,Flags)
*
* Description:	Close a previously opened file. No futher file operations,
*					including SELECT or READNEXT will be possible.
*
* Parameters:	FileDescriptor:	The address of the file descriptor that
*								was created when the file was originally opened.
*				Flags:			Can be one or more of the following bit patterns :
*									None supported at the moment.
*
* Return Value:	0 means all went okay, any other value is the errno reason
*					for failure.
*
* Notes:		All locks associated with this file descriptor will be released.
*
*******************************************************************/

static	int		ExampleCLOSE(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor, int Flags)
{
	struct	TypeExample		* typeptr = FileDescriptor->TypePtr ;
	/*
	 * Close the file -- this also releases all the locks.
	 */
	JRunFileIOClose(dp , &FileDescriptor->fdinfo , 1 );
	/*
	 * Delete all workspace WE may have allocated.
	 */
	/* 
	 * Delete the workspaces used for this opened file descriptor.
	 */
#ifdef WIN32
/*
 * On windows, a dll must release the space that it allocated itself 
 * (The jedi routines use a different heap), otherwise you get an
 * Debug assertion error. 
 */
	free((void *)typeptr->pathname);
#else
	if (typeptr->pathname != NULL)
	{
		free(typeptr->pathname);
	}
	JediFreeFileDescriptor(dp,&FileDescriptor);				/* Free all space allocated by this file descriptor */
#endif

	return 0 ;

}

/**************************************************************
*
* Function:		ExampleSELECT(FileDescriptor , SelectPtr )
*
* Description:	Perform the SELECT command on a Unix directory.
*
* Parameters:	FileDescriptor:	The address of the file descriptor that
*								was created when the file was originally opened.
*				SelectPtr:		The address of a select pointer that will be 
*								updated if the select was successful.
*
* Return Value:	0 means all went okay, any other value is the errno reason
*					for failure.
*
* Notes:		None.
*
*******************************************************************/

static	int		ExampleSELECT(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor ,
								struct JediSelectPtr ** SelectPtr )
{

	DIR		*Dir;
	struct	JediSelectPtr	*BlockPtr , **LoopPtr;
	struct	TypeExample		* typeptr = FileDescriptor->TypePtr ;

	/*
	 * Open the directory using opendir() so we can select all records.
	 */
	if ((Dir = opendir(typeptr->pathname)) == NULL) 
	{
		return errno ;
	}
	/* 
	 * Allocate space for our own control block and the normal select control block.
	 */
	if ((BlockPtr=malloc(sizeof(struct JediSelectPtr))) == NULL) 
	{
		return errno ; 
	}
	if ((BlockPtr->TypePtr=malloc(sizeof(struct UnixSelect))) == NULL) 
	{
		return errno ;
	}
	((struct UnixSelect *)BlockPtr->TypePtr)->SelectDIR = Dir;		/* Update the pointer with our select using opendir() */
	/*
	 * Now to add our newly allocated space onto the end of the Select chain, as pointed
	 * to by the file descriptor.
	 */
	LoopPtr = &FileDescriptor->SelectPtr;				/* Point to the first pointer */
	while(*LoopPtr != NULL) 
	{
		LoopPtr = (struct JediSelectPtr **)(*LoopPtr);	/* Get address of next pointer */
	}
	*LoopPtr = BlockPtr ;								/* Update at the end of the chain */
	BlockPtr->NextPtr = NULL;							/* As we are the last in the chain, null the next pointer */
	/* 
	 * Give user the address of the structure we have allocated.
	 */
	*SelectPtr = BlockPtr ;
	return 0 ;											/* Return with good completion code */
}

/**************************************************************
*
* Function:		ExampleSELECTEND(FileDescriptor , SelectPtr)
*
* Description:	Terminate the SELECT command on a Unix directory.
*
* Parameters:	FileDescriptor:	The address of the file descriptor that
*								was created when the file was originally opened.
*				SelectPtr:		A select pointer.
*
* Return Value:	0 means all went okay, any other value is the errno reason
*					for failure.
*
* Notes:		None.
*
*******************************************************************/

static	int		ExampleSELECTEND(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor ,
								struct JediSelectPtr * SelectPtr)
{

	struct	JediSelectPtr	*Next , ** LoopPtr ;

	closedir(((struct UnixSelect *)SelectPtr->TypePtr)->SelectDIR);	/* Close the current directory */
	free(SelectPtr->TypePtr);				/* Free the first lot of workspace */
	Next = SelectPtr->NextPtr ;							/* Save any link information */
	free(SelectPtr);						/* Free the second lot of workspace */

	/*
	 * The structure passed to us is part of a linked list of workspaces that
	 * begin at address FileDescriptor->SelectPtr. The address of our workspace is given
	 * by the variable 'SelectPtr' and the forward pointer for our linked workspace is
	 * given by the variable 'Next'. We now have to go through the list and remove our
	 * workspace from the linked list.
	 */
	LoopPtr = &FileDescriptor->SelectPtr;				/* Point to the first pointer */
	for(;;) 
	{
		if (*LoopPtr == NULL) break;					/* This should never happen !!! */
		if (*LoopPtr == SelectPtr) 						/* We have found a forward link to our select list */
		{
			*LoopPtr = Next ;							/* This effectively removes our select list from the linked list */
			break;
		}
		LoopPtr = (struct JediSelectPtr **)(*LoopPtr);	/* Get address of next pointer */
	}

	return 0 ;											/* Return with good completion code */

}

/**************************************************************
*
* Function:		ExampleREADNEXT(FileDescriptor , SelectPtr , RecordKeyPtr , RecordLenPtr)
*
* Description:	Perform the REANDEXT command on a Unix directory. This reads the next
*				record key (i.e. Unix file name) from the directory.
*
* Parameters:	FileDescriptor:	The address of the file descriptor that
*								was created when the file was originally opened.
*				SelectPtr:		The address of a select pointer that was created during
*								a SELECT command.
*				RecordKeyPtr:	Pointer to a char pointer showing the address of
*								the default buffer to use. If this buffer is not big enough
*								(see RecordLenPtr) then we use malloc() to create
*								a workspace big enough.
*				RecordKeyLen:	Pointer to an integer that initially describes the
*								length of the default buffer supplied by RecordKeyPtr . We
*								also return here a value which shows the length of the next 
*								record key , or (-1) if we come to the end of the list.
*
* Return Value:	0 means all went okay, any other value is the errno reason
*					for failure.
*
* Notes:		If we detect an error, or we come to the end of the
*					list of record keys, then we will NOT remove the select
*					list structure from the linked list of select lists
*					defined by the file descriptor. It is up to the calling
*					application to use the JediSelectEnd function call
*					to release the space associated with a select list block.
*				For application ease, we suppress the return of the special
*					cases '.' and '..'. This is because the application will want to
*					usually see just normal files in a directory, and not '.' ones.
*					We COULD if we wanted to suppress all hidden files that begin 
*					with '.', but then we have to mess about with extra commands so
*					they could actually read it if the wanted. 
*
*******************************************************************/

static	int		ExampleREADNEXT(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor ,
								struct JediSelectPtr * SelectPtr , char ** RecordKeyPtr , int * RecordLenPtr )
{

	struct	dirent	*DirPtr;
	int		i1;

	errno = 0;

	for(;;) 
	{
		/*
		 * Use a call to readdir() to get the next Unix file name in the directory.
		 */
		if ((DirPtr = readdir(((struct UnixSelect *)SelectPtr->TypePtr)->SelectDIR)) == NULL) 	/* Cannot read any more */
		{
			(*(RecordLenPtr)) = (-1);					/* Show there is no more record keys */
			return errno ;
		}

		if (DirPtr->d_name[0] == '.') 					/* See if a special case */
		{
			if (strcmp(DirPtr->d_name,".") == 0) continue;
			if (strcmp(DirPtr->d_name,"..") == 0) continue;
		}
		break;
	}

	i1 = strlen(DirPtr->d_name);						/* Get the length of the next file name */

	if (*RecordLenPtr < i1) 							/* See if enough space to store the record key */
	{
		if ((*RecordKeyPtr = malloc(i1)) == NULL) /* Allocate enough space, or check for errors */
		{
			return errno;
		}
	}
	*RecordLenPtr = i1;									/* Return to user the length of the record key */
	memmove(*RecordKeyPtr,DirPtr->d_name,i1);		/* Copy the name into the users address space */
			
	return 0 ;											/* Return with good completion code */

}

/**************************************************************
*
* Function:		ExampleREADRECORD(FileDescriptor , Flags , RecordKey , RecordKeyLen ,
*					BufferPtr , BufferLenPtr , FieldNumber , mallocptr )
*
* Description:	Read a record from a Unix directory.
*
* Parameters:	FileDescriptor:	The address of the file descriptor that
*								was created when the file was originally opened.
*				Flags:			One of the following OR'ed together
*					JEDI_RECORD_FIELDREADWRITE		The FieldNumber value shows to extract a single field from
*													the record. This is only supported if the JEDI_STATUS_FIELDREAD
*													bit is set in the Status field of the file descriptor.
*				BufferPtr:		Pointer to a char pointer showing the address of
*								the default buffer to use to read into. If this buffer is not big enough
*								(see RecordLenPtr) then we use malloc() to create
*								a workspace big enough.
*				BufferLenPtr:	Pointer to an integer that initially describes the
*								length of the default buffer supplied by BufferPtr . We
*								also return here a value which shows the length of the
*								record we have read in.
*				FieldNumber:	If the JEDI_RECORD_FIELDREADWRITE bit is set, it
*								shows to return this field only.
*				mallocptr:		If we need to resize the buffer at 'BufferPtr', then we call this
*								function as a replacement for malloc(). This is simply a
*								performance tuning exercise. You would normally pass malloc here.
*
* Return Value:	0		The record was read in okay and locks (if necessary) we taken.
*				ENOENT	No such record exists
*				JEDI_ERRNO_LOCK_TAKEN	The record exists, but
*				EDEADLK	One of the lock bits was set, and a deadly embrace situation was
*						detected and avoided.
*				Any other value is an undeterminate error taken from the errno value
*					passed back by a Unix system function call.
*
*				The value at FileDescriptor->ConvertFlags has the bit JEDI_CONVERT_READ_BINARY
*					set if the operation performed was a BINARY read (i.e. no conversions), or
*					the bit is reset if some sort of conversion took place.
*
* Notes:		We may convert the record key to become a Unix one.
*
*******************************************************************/

static	int		ExampleREADRECORD(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags , char * RecordKey ,
									int RecordKeyLen , char ** BufferPtr , int * BufferLenPtr , int FieldNumber ,
									void *(mallocptr)() )
{

	char	ActualPathName[PATH_MAX+1];
	struct	stat StatInfo;
	int		SaveErrno ;
	char	*BufferAddress;
	int		RecordSize , i1 , fd1 , ReadBinary ;
	unsigned char *Ptr1 , *Ptr2 ;
	/*
	 * Work out a full path name to use.
	 */
	if ((errno=JediUnixMakeFullName(dp , FileDescriptor,RecordKey,RecordKeyLen,ActualPathName,sizeof(ActualPathName))) != 0) 
	{
		return errno ;
	}
	/* 
	 * Open file , exit loop if all went okay.
	 */
	if ((fd1=open(ActualPathName,O_RDONLY| O_BINARY)) < 0) 
	{
		return errno ;	
	}
	/* 
	 * Get file statistics.
	 */
	if (fstat(fd1,&StatInfo) != 0)
	{
		SaveErrno = errno ;
		close(fd1);
		return SaveErrno ;
	}

	RecordSize = StatInfo.st_size;						/* Return to user the size of the record */
		
	BufferAddress = (*BufferPtr);						/* Get address of the start of the buffer */
	if (RecordSize > 0) 								/* Make sure something there to read in */
	{
		if (RecordSize > (*BufferLenPtr)) 				/* Not enough space in the users buffer */
		{
			if ((BufferAddress = JediBaseReadMalloc(dp , Flags , mallocptr , BufferPtr , RecordSize)) == NULL)
			{
				SaveErrno = errno;						/* Could get corrupted by the close() function */
				close(fd1);
				return SaveErrno ;
			}
		} 
		if ((i1=read(fd1,BufferAddress,RecordSize)) != RecordSize) /* Read in number of bytes requested */
		{
			SaveErrno = errno;							/* Could get corrupted by the close() function */
			close(fd1);									/* So not need the file any more so close it */
			return(SaveErrno==0 ? EIO : SaveErrno);		/* Return with error condition */
		}
	}
	close(fd1);											/* Do not need the file any more so close it */

	/*
	 * The following code is concerned with conversion of new-lines to attribute marks.
	 * A Unix source file (e.g. text file) will normally have lines of source delimited
	 * by new-line characters, and the file will have a trailing new-line. 
	 * A Unix binary file (e.g. terminfo database) will normally be recognised as such by
	 * having the last character NOT being a new-line.
	 *
	 * The default action is to look at the last character to see if it is a text or binary record.
	 * If a binary record, the default action is to return the record exactly as is.
	 * If a text record, the default action is to strip the trailing new-line and convert
	 *    the remaining new-lines into attribute marks (0xfe).
	 * These default actions can be over-ruled by calls to the ExternalIOCTL() function.
	 */
	if (RecordSize == 0 ) 
	{
		ReadBinary = ((FileDescriptor->ConvertFlags & (JEDI_CONVERT_READ_TEXT|JEDI_CONVERT_READ_INTELLIGENT)) ? 0 : JEDI_CONVERT_LAST_READ_BINARY) ;
	} else {
		ReadBinary = JEDI_CONVERT_LAST_READ_BINARY;		/* Default to showing that NO conversion took place */
		if ((FileDescriptor->ConvertFlags & (JEDI_CONVERT_READ_TEXT|JEDI_CONVERT_READ_INTELLIGENT)) && S_ISREG(StatInfo.st_mode )) /* We can only do the conversion on regular files */
		{
			if (FileDescriptor->ConvertFlags&JEDI_CONVERT_READ_INTELLIGENT)	/* We may want to do some conversion  - decide */
			{
				if (*(BufferAddress+RecordSize-1) == '\n') /* If the last character is a new-line, it is usually a text file */
				{
					RecordSize-- ;						/* Decrement the record size */
					ReadBinary = 0;
				}
			} else {									/* We always want to do some conversion */
				if (*(BufferAddress+RecordSize-1) == '\n') 
				{
					RecordSize--;						/* Decrement the record size */
				}
				ReadBinary = 0;
			}
		}

		/* 
		 * Convert to binary if necessary.
		 */
		if (!(ReadBinary)) 						/* See if finally to do any conversions */
		{
			i1 = RecordSize ; Ptr1 = (unsigned char *)BufferAddress;		/* Initialise before we look for new-line characters */
			while(i1) 
			{
				if ((Ptr2=memchr(Ptr1,'\n',i1))) 		/* Find the next new-line delimiter */
				{
					*((unsigned char *)Ptr2++) = 0xfe;	/* Convert new-line delimiter to attribute-mark delimiter */
					i1 -= (Ptr2-Ptr1);					/* Adjust number of bytes remaining */
					Ptr1 = Ptr2;
				} else i1 = 0;
			}
		}
	}

	*BufferLenPtr = RecordSize ;						/* Let the user know how much space we used */

	return 0 ;											/* Return with good completion code */
}

/**************************************************************
*
* Function:		ExampleWRITERECORD(FileDescriptor , Flags , RecordKey , RecordKeyLen ,
*					BufferPtr , BufferLen , FieldNumber)
*
* Description:	Write a record to a Unix directory.
*
* Parameters:	FileDescriptor:	The address of the file descriptor that
*								was created when the file was originally opened.
*				Flags:			One of the following OR'ed together
*					JEDI_RECORD_FIELDREADWRITE		The FieldNumber value shows to write a single field from
*													the record. This is only supported if the JEDI_STATUS_FIELDWRITE
*													bit is set in the Status field of the file descriptor. For the
*													moment, we will not support it. Maybe we can do it later.
*					JEDI_RECORD_SCRATCHWRITE		Shows if the buffer passed by BufferPtr and BufferLen is a
*													scratch area than can be overwritten.
*				BufferPtr:		Pointer to a char pointer showing the address of the data to write. See also the
*								notes on JEDI_RECORD_SCRATCHWRITE above.
*				BufferLen:		The amount of data to write.
*				FieldNumber:	If the JEDI_RECORD_FIELDREADWRITE bit is set, it shows to return this field only.
*
* Return Value:	0		The record was read in okay and locks (if necessary) we taken.
*				ENOENT	No such record exists
*				JEDI_ERRNO_LOCK_TAKEN	The record exists, but
*				EDEADLK	One of the lock bits was set, and a deadly embrace situation was
*						detected and avoided.
*				Any other value is an undeterminate error taken from the errno value
*					passed back by a Unix system function call.
*
*				The value at FileDescriptor->ConvertFlags has the bit JEDI_CONVERT_WRITE_BINARY
*					set if the operation performed was a BINARY read (i.e. no conversions), or
*					the bit is reset if some sort of conversion took place.
*
* Notes:		None.
*
*******************************************************************/

static	int		ExampleWRITERECORD(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags , 
									char * RecordKey , int RecordKeyLen , char * BufferPtr , int BufferLen , int FieldNumber)
{

	char	ActualPathName[PATH_MAX+1];
	unsigned char *Ptr1 , *Ptr2 , *Ptr3 , *WriteAddr ;
	int		i1 , i2 , WriteActual , fd1  , SaveErrno ;

	if ((errno=JediUnixMakeFullName(dp,FileDescriptor,RecordKey,RecordKeyLen,ActualPathName,sizeof(ActualPathName))) != 0) 
	{
		return errno ;
	}

	if ((fd1=open(ActualPathName,O_WRONLY|O_TRUNC|O_CREAT|O_BINARY,0640)) < 0) 	/* Open file , exit loop if all went okay */
	{
		return errno;
	}

	if (BufferLen > 0) 										/* Make sure something there to write to */
	{
		/*
		 * Now to write the actual data. Before we do this, we have to decide if to write back
		 * completely as-is , or to translate from BASIC delimiters to Unix file delimiters.
		 * Note at this point the bit FileDescriptor->Status&JEDI_RECORD_SCRATCHWRITE, which shows if the
		 * data buffer passed is a scratch buffer that can be corrupted.
		 * 
		 * If the FileDescriptor->ConvertFlags & JEDI_CONVERT_WRITE_TEXT bit is set, then we ALWAYS do the conversion.
		 * If the FileDescriptor->ConvertFlags & JEDI_CONVERT_WRITE_BINARY bit is set, then we NEVER do the conversion.
		 * If the FileDescriptor->ConvertFlags & JEDI_CONVERT_WRITE_DECIDE bit is set, we do the conversion assuming
		 *    the last character is NOT a new-line.
		 */
		if (!(FileDescriptor->ConvertFlags&JEDI_CONVERT_WRITE_BINARY) && (FileDescriptor->ConvertFlags&JEDI_CONVERT_WRITE_TEXT || (FileDescriptor->ConvertFlags&JEDI_CONVERT_WRITE_INTELLIGENT && (*(BufferPtr+BufferLen-1) != '\n'))))
		{
			/*	
			 * We have decided to do a conversion. We may be able to do the conversion in-house, 
			 * or we may have to allocate an external buffer.
			 */
			if (Flags&JEDI_RECORD_SCRATCHWRITE) 		/* We can re-use the space */
			{
				Ptr1 = WriteAddr = (unsigned char *)BufferPtr;	/* Set up to show where to convert from */
				i1 = BufferLen;
				while(i1 && (Ptr2=memchr(Ptr1,0xfe,i1)))/* Look for next new-line delimiter */
				{
					*(Ptr2++) = '\n';					/* Convert attribute mark delimiter to new-line delimiter */
					i1 -= (Ptr2-Ptr1);					/* Adjust number of bytes left */
					Ptr1 = Ptr2;						/* Point Ptr1 to the next character to start scanning at */
				}
			} else {
				/*	
				 * We need to convert , must allocate some scratch space. If we get an
				 * error, it is so catastrophic we need not bother tidying up.
				 */
				if ((Ptr1 = WriteAddr = malloc(BufferLen)) == NULL) /* Allocate using user-defined function */
				{
					return errno;
				}
				i1 = BufferLen;							/* Number of bytes to copy */
				Ptr2 = (unsigned char *)BufferPtr;		/* Address to copy FROM */
				while(i1) 
				{
					if ((Ptr3=(unsigned char *)memccpy(Ptr1,Ptr2,0xfe,i1)) == NULL) break;	/* Copy until attribute-mark was seen */
					i2 = (Ptr3-Ptr1);					/* Number of bytes copied in above memccpy() */
					Ptr2 += i2;							/* Update the source for the next copy by how many bytes were written */
					Ptr1 += i2;							/* Update the target for the next copy by how many bytes were written */
					i1 -= i2;							/* Update the number of bytes remaining to be copied */
					*(Ptr3-1) = '\n';					/* Replace attribute mark delimiter with new-line delimiter */
				}
			}
			WriteActual = write(fd1,WriteAddr,BufferLen);	/* Write out the converted data */
			if (*(WriteAddr+BufferLen-1) != '\n') 			/* In text mode, we MUST terminate with a new-line */
			{
				WriteActual += write(fd1,"\n",1);			/* Write out a trailing new-line */
				BufferLen++;								/* Update expected number of bytes to have been written */
			}
			if (WriteAddr != (unsigned char *)BufferPtr) 	/* Release space if it had beed malloc'ed */
			{
				free(WriteAddr);	
			}
		} else {
			WriteActual = write(fd1,BufferPtr,BufferLen);	/* Write the data out */
		}
		if (WriteActual != BufferLen) 					/* Make sure the write was successful !! */
		{
			SaveErrno = errno;							/* Could get corrupted by the close() function */
			close(fd1);									/* So not need the file any more so close it */
			return(SaveErrno==0 ? EIO : SaveErrno);		/* Return with error condition */
		}
	}
	close(fd1);											/* Do not need the file any more so close it */

	return 0 ;											/* Return with good completion code */
}

/**************************************************************
*
* Function:		ExampleDELETERECORD(FileDescriptor , Flags , RecordKey , RecordKeyLen )
*
* Description:	Delete a record from a file.
*
* Parameters:	FileDescriptor:	The address of the file descriptor that
*								was created when the file was originally opened.
*				Flags:			The following bits can be OR'ed together -
*								-- None used so far , the caller should always pass 0 --
*				RecordKey:		Address of the record key describing the record to delete.
*				RecordKeyLen:	Length of the record key.
*
* Return Value:	0		Record deleted okay.
*				ENOENT	No record there to delete.
*
* Notes:		None.
*
*******************************************************************/

static	int		ExampleDELETERECORD(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags ,
									char * RecordKey , int RecordKeyLen )
{

	char	ActualPathName[PATH_MAX+1];

	if ((errno=JediUnixMakeFullName(dp , FileDescriptor,RecordKey,RecordKeyLen,ActualPathName,sizeof(ActualPathName)))==0)
	{
		if (!unlink(ActualPathName))
		{
			errno=0;								/* Actually delete the file, set errno if it fails */
		}
	}

	return errno ;

}

/**************************************************************
*
* Function:		ExampleLOCK(FileDescriptor , Flags , RecordKey ,
*					RecordKeyLen )
*
* Description:	Set or release a lock from a given file.
*				Basically, we use the jBASE supplied locking mechanism.
*
* Parameters:	FileDescriptor:	The address of the file descriptor that
*								was created when the file was originally opened.
*				Flags:			One of the following OR'ed together
*					JEDI_RECORD_LOCKRECORD			Take a record lock and wait.
*					JEDI_RECORD_LOCKRECORD_NOWAIT	Take a record lock , but return if lock already taken .
*					JEDI_RECORD_UNLOCKRECORD		Remove a record lock
*				RecordKey:		Address of the record key describing the record lock to
*								take. If this value is NULL and we are releasing record locks, then
*								we release ALL record locks for the file (but NOT the file lock).
*				RecordKeyLen:	Length of the record key.
*
* Return Value:	0		The locks were created/removed okay.
*				JEDI_ERRNO_LOCK_TAKEN	The JEDI_RECORD_LOCKRECORD_NOWAIT bit
*						was set, and the record/field was already locked by another process.
*				EDEADLK	One of the lock bits was set, and a deadly embrace situation was
*						detected and avoided.
*				Any other value is an undeterminate error taken from the errno value
*					passed back by a Unix system function call.
*
*
* Notes:		None.
*
*******************************************************************/

static	int		ExampleLOCK(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags , 
							char * RecordKey , int RecordKeyLen )
{
	return JRunFileIOLockRecord(dp , FileDescriptor->fdinfo , Flags , RecordKey , RecordKeyLen );

}

/**************************************************************
*
* Function:		ExampleIOCTL(FileDescriptor , SubCommand, IoctlAddr , IoctlLen 
*					IoctlReturnAddr , IoctlReturnLen)
*
* Description:	Various miscellaneous file control functions.
*				The sub-functions in here can be left out if necessary, as IOCTL is
*				mostly device dependent. All of the jBASE utilities that use IOCTL()
*				calls will work correctly without any user-written database drivers
*				having identical calls.
*
* Parameters:	FileDescriptor:		The address of the file descriptor that
*									was created when the file was originally opened.
*				SubCommand:			The sub-command to perform
*				IoctlAddr:			Address of structure to supplement the command.
*									This value can be NULL.
*				IoctlLen:			Length of structure to supplement the command.
*				IoctlReturnAddr:	Address of where to return any data.
*									This value can be NULL.
*				IoctlReturnLen:		Length of IoctlReturnAddr - Also gets
*									returned with actual amount of data returned at IoctlReturnAddr.
*									This value can be NULL, but only if IoctlReturnAddr is also NULL.
*
* Return Value:	0		IOCTL command executed okay. Any other value is
*						the errno reason.
*
* Notes:		None.
*
*******************************************************************/

static	int		ExampleIOCTL(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int SubCommand ,
							void * IoctlAddr , int IoctlLen ,void * IoctlReturnAddr , int * IoctlReturnLen )
{

	int				ReturnValue = 0 ;
	int				i1 , ReturnBufferLen ;
	char			ReturnBuffer[4096];
	char			ActualPathName[PATH_MAX+1];
	int				fd1 ;

	switch(SubCommand) 
	{

	case	JEDI_IOCTL_CONVERT:							
		/* 
		 * Convert from one value to another.
		 */
		if (IoctlReturnAddr != NULL && IoctlAddr != NULL) 	/* Make sure something to return */
		{
			FileDescriptor->ConvertFlags = JediBaseIoctlConvert(dp , FileDescriptor->ConvertFlags , IoctlAddr , IoctlLen , IoctlReturnAddr , IoctlReturnLen) ;
		}
		break;

	case	JEDI_IOCTL_FILESTATUS:						
		/*	
		 * Details of the file.
		 *
		 * Return details as an attribute delimited record as follows :
		 * <1>		File Type as a string
		 * <2>		FileFlags , as decimal number , showing the LOG, BACKUP and TRANSACTION permissions.
		 * <3>		BucketQty , as decimal number , number of buckets in the file.
		 * <4>		BucketSize , as decimal number , size of each bucket in bytes.
		 * <5>		SecSize , as decimal number , size of secondary data space.
		 * <6>		Restore Spec , a string showing any restore re-size specification.
		 * <7>		Locking identifiers, delimited by multi-values.
		 * <8>		FileFlags  showing LOG, BACKUP and TRANSACTION permissions. Multi-value is is
		 * 			<8,1> set to 1 if no logging , <8,2> set to 1 if no transaction and <8,3> set to 1 if no file backup.
		 */
		ReturnBufferLen = sprintf(ReturnBuffer,"Example\3760\376\376\376\376\376%d\375%d\3760\3750\3750", JRunFileIOGetLock1(dp , FileDescriptor->fdinfo),JRunFileIOGetLock2(dp , FileDescriptor->fdinfo));

		if (IoctlReturnAddr != NULL) 							/* Make sure something to return */
		{
			i1 = min(ReturnBufferLen,*IoctlReturnLen);			/* Amount of data to copy */
			memmove(IoctlReturnAddr,ReturnBuffer,i1);		/* Copy back exactly what seen */
			*IoctlReturnLen = i1 ;
		} else {
			ReturnValue = EINVAL;
		}
		break;

	case JEDI_IOCTL_FINDRECORD:			
		/* 
		 * Test whether record exists.
		 * Work out a full path name to use.
		 */
		if ((ReturnValue=JediUnixMakeFullName(dp , FileDescriptor,IoctlAddr,IoctlLen,ActualPathName,sizeof(ActualPathName))) == 0) 
		{
			if ((fd1=open(ActualPathName,O_RDONLY|O_BINARY)) < 0) 
			{
				ReturnValue = errno;
			}
			close(fd1);
		}
		/* 
		 * Check ReturnAddr for NULL before returning value.
		 */
		if ( IoctlReturnAddr != NULL )
		{
			i1 = min(IoctlLen,*IoctlReturnLen);			/* Amount of data to copy */
			memmove(IoctlReturnAddr,IoctlAddr,i1);	/*Copy back source variable*/
			*IoctlReturnLen = i1 ;
		}

		break;

	default:											
		/* 
		 * No action necessary on non-specific actions.
		 */
		if (IoctlReturnAddr != NULL && IoctlAddr != NULL) 	/* Make sure something to return */
		{
			i1 = min(IoctlLen,*IoctlReturnLen);				/* Amount of data to copy */
			memmove(IoctlReturnAddr,IoctlAddr,i1);		/* Copy back exactly what seen */
			*IoctlReturnLen = i1 ;
		}
		ReturnValue = -1;
		break;
	}

	return ReturnValue ;

}

/**************************************************************
*
* Function:		ExampleCLEARFILE(FileDescriptor )
*
* Description:	Delete all records in the given file.
*
* Parameters:	FileDescriptor:	The address of the file descriptor that
*								was created when the file was originally opened.
*
* Return Value:	0 means all went okay, any other value is the errno reason
*					for failure.
*
* Notes:		If we get any errors during the unlinking of files, then
*					we do NOT report an error. Such errors may be invalid
*					permissions, trying to delete a directory etc.
*
*******************************************************************/

static	int		ExampleCLEARFILE(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor )
{
	DIR						* DirPtr;
	struct	TypeExample		* typeptr = FileDescriptor->TypePtr ;
	int						ReturnValue = 0 ;
	struct	dirent			* Next;
	char					PathName[PATH_MAX*2];
	struct	stat 			stat_info ;

	if ((DirPtr = opendir(typeptr->pathname)) == NULL) 		/* Perform the SELECT on the directory */
	{
		return errno;
	}

	while ((Next = readdir(DirPtr)) != NULL) 				/* Get next directory entry name */
	{
		if (JLibSignalTest(dp))
		{
			ReturnValue = EINTR ;
			break ;
		}
		strcpy(PathName,typeptr->pathname);					/* Copy the directory name */
		strcat(PathName,DIR_DELIM_STR);						/* Add a delimiter */
		strcat(PathName,Next->d_name);						/* Copy the file name within the directory */
		/*
		 * Find out the type of the entry. We will not attempt
		 * to delete anything other than regular files. For example, under AIX,
		 * a root user doing a CLEAR-FILE can delete . and .., thus causing all
		 * sorts of lost space and problems.
		 */
		if (lstat(PathName , &stat_info) == 0)
		{
			if (S_ISREG(stat_info.st_mode))
			{
				unlink(PathName);								/* Delete the file, ignore any errors */
			}
		}
	}

	closedir(DirPtr);									/* Close the unix directory */

	return ReturnValue;

}

/**************************************************************
*
* Function:		ExampleDELETEFILE(FileDescriptor , Flags )
*
* Description:	Delete a file
*
* Parameters:	FileDescriptor  - The address of the file descriptor that
*					was created when the file was originally opened.
*				Flags - Miscellaneous flags .
*
* Return Value:	0 means all went okay, any other value is the errno reason
*					for failure.
*
* Notes:		If we get any errors during the unlinking of files, then
*					we do NOT report an error. Such errors may be invalid
*					permissions, trying to delete a directory etc.
*
*******************************************************************/

static	int		ExampleDELETEFILE(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor , int Flags )
{
	int		ReturnCode ;
	char	RmCommand[PATH_MAX+20] ;
	char	PathName[PATH_MAX+20] ;
	char	FilePathName[PATH_MAX+20] ;
	int		i1 ;
	struct	TypeExample		* typeptr = FileDescriptor->TypePtr ;
	/*	
	 * Build up a delete command for the removal of the last directory.
	 * There are two to build up--the real directory and the 
	 * file pointer.
	 */
	strcpy(PathName,typeptr->pathname);
	strcpy(FilePathName,PathName);
	i1 = strlen(FilePathName);
	if (i1 >= 2 && memcmp(&FilePathName[i1-2],".d",2) == 0)
	{
		FilePathName[i1-2] = '\0';
	}
	else
	{
		FilePathName[0] = '\0';
	}
	sprintf(RmCommand,"rm -rf %s",PathName);
	/*
	 * Close the file passed.
	 */
	JediClose(dp , FileDescriptor , 0 );
	/*	
	 * Issue the command to recursively delete the directory .
	 */
	if (!(ReturnCode=system(RmCommand))) {
		/*	
		 * Delete the sub-directories - do not worry about return codes.
		 */
		JediBaseDeleteSubDirectories(PathName , Flags );
	}
	/*
	 * Delete the file path name -- don't worry about return codes.
	 */
	sprintf(RmCommand,"rm -f %s",FilePathName);
	system(RmCommand);
	/*
	 * Return with correct return code.
	 */
	return ReturnCode ;	
}

/***************************************************************
*
* Function:		ExampleSYNC(FileDescriptor)
*
* Description:	Synchronise all file I/O with disk i.e. flush output.
*
* Parameters:	FileDescriptor:	The opened file descriptor.
*
* Return Value:	0 means all worked okay, otherwise the errno reason is returned.
*
* Notes:		None.
*
*******************************************************************/

static	int		ExampleSYNC(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor)
{
	/* 
	 * Not really supported ... 
	 */
	return 0 ;
}

/******************************************************************
*
* Function:		JediUnixMakeFullName(FileDescriptor,RecordKey,RecordKeyLen,
*					NewAddr , NewLen)
*
* Description:	Create a full path name, given the directory and record name
*
* Parameters:	FileDescriptor:	Address of the file descriptor returned during
*								the open() command.
*				RecordKey:		Address of the record key (i.e. file name). 
*				RecordKeyLen:	Length of above record key
*				NewAddr:		Address of buffer where to store the result as a
*								NULL terminated string.
*				NewLen			Length of above buffer, including the 1 byte needed for
*								the NULL terminator.
*
* Return Value:	0 means all went okay. E2BIG means the buffer was not big enough.
*
* Notes:		None.
*
****************************************************************************/

static	int		JediUnixMakeFullName(struct jBASEDataAreas * dp , JediFileDescriptor * FileDescriptor, char * RecordKey,
									int RecordKeyLen, char * NewAddr, int NewLen)
{
	struct	TypeExample		* typeptr = FileDescriptor->TypePtr ;
	/*	
	 * Check validity of size of combined string .
	 */
	if (typeptr->pathnamelen+RecordKeyLen+2 > NewLen) 
	{
		return E2BIG ;
	}
	/*	
	 * Copy the absolute path name to start off with.
	 */
	memmove(NewAddr,typeptr->pathname,typeptr->pathnamelen);
	NewAddr += typeptr->pathnamelen;
	/*	
	 * Add a '/' as a file delimiter.
	 */
	*(NewAddr++) = DIR_DELIM_CH;
	/*	
	 * Add the file name to the directory path name.
	 */
	memmove(NewAddr,RecordKey,RecordKeyLen);
	/*	
	 * Make the result a NULL terminated string.
	 */
	*(NewAddr+RecordKeyLen) = '\0' ;

	return 0 ;
}
